﻿using System.Collections.Generic;
using BNFUP.Rules;

namespace BNFUP.Interfaces
{
    /// <summary>
    /// This interface must be implemented by all the objects that participate in the compiler work
    /// </summary>
    public interface IRuleItem
    {
        /// <summary>
        /// Indicates that all the events in the object had been linked
        /// </summary>
        bool Completed { get; set; }
        /// <summary>
        /// Indicates if the item is mandatory or optional
        /// </summary>
        bool Optional { get; set; }
        /// <summary>
        /// Description of the item
        /// </summary>
        string Description { get; }
        /// <summary>
        /// Enumerates the child items, if any
        /// </summary>
        IEnumerable<IRuleItem> Childs { get; }
        /// <summary>
        /// Returns the source text corresponding to this item
        /// </summary>
        /// <returns>Text extracted from the source code</returns>
        string GetText();
        /// <summary>
        /// Used during construction, replace temporary rules with the definitive ones
        /// </summary>
        /// <param name="name">Rule name</param>
        /// <param name="rule">Definitive version of the rule</param>
        /// <param name="current">Used internally to avoid recursive loops, pass an empty string</param>
        void UpdateRule(string name, RuleBase rule, string current);
        /// <summary>
        /// Change a rule with a new version
        /// </summary>
        /// <param name="oldrule">Old version of the rule</param>
        /// <param name="newrule">New version of the rule</param>
        /// <param name="current">Used internally to avoid recursive loops, pass an empty string</param>
        void ChangeRule(RuleBase oldrule, RuleBase newrule, string current);
        /// <summary>
        /// Remove all the references to the provided rule
        /// </summary>
        /// <param name="rule">Rule to remove</param>
        /// <param name="current">Used internally to avoid recursive loops</param>
        void DeleteRule(RuleBase rule, string current);
        /// <summary>
        /// Test if the rule matches at the current parser position
        /// </summary>
        /// <param name="parser">Parser with the source code</param>
        /// <returns>True if the rule matches at current position</returns>
        bool Matches(Parser parser);
        /// <summary>
        /// Try to build the item using the current source code position
        /// </summary>
        /// <param name="parser">Parser with the source code</param>
        void Build(Parser parser);
    }
}
