﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text.RegularExpressions;
using BNFUP.Properties;
using BNFUP.Exceptions;

namespace BNFUP
{
    public class Parser
    {
        private int _line;
        private int _position;
        private string _data;
        private Stack<KeyValuePair<int, int>> _bookmarks = new Stack<KeyValuePair<int, int>>();
        public Parser(TextReader rdr)
        {
            _data = rdr.ReadToEnd();
            if (string.IsNullOrEmpty(_data))
            {
                throw new EndOfCodeException();
            }
            _line = 1;
            _position = 0;
            SkipWhitespaces();
            if (EOF)
            {
                throw new EndOfCodeException();
            }
        }

        public bool CaseSensitive { get; set; }
        public bool EOF
        {
            get
            {
                return _data.Length <= _position;
            }
        }
        public void SavePosition()
        {
            _bookmarks.Push(new KeyValuePair<int, int>(_line, _position));
        }
        public void RestorePosition()
        {
            KeyValuePair<int, int> kv = _bookmarks.Pop();
            _line = kv.Key;
            _position = kv.Value;
        }
        public string DiscardSavePoint()
        {
            KeyValuePair<int, int> pi = _bookmarks.Pop();
            string s = _data.Substring(pi.Value, _position - pi.Value);
            if (!string.IsNullOrEmpty(s))
            {
                while (char.IsWhiteSpace(s[s.Length - 1]))
                {
                    s = s.Substring(0, s.Length - 1);
                }
            }
            return s;
        }
        public bool TryParse(string token)
        {
            SkipWhitespaces();
            if (EOF)
            {
                return false;
            }
            return _data.Substring(_position).StartsWith(token, CaseSensitive ? StringComparison.Ordinal : StringComparison.OrdinalIgnoreCase);
        }
        public bool TryRegex(string rex)
        {
            if (EOF)
            {
                return false;
            }
            Regex reg = new Regex(rex, CaseSensitive ? RegexOptions.None : RegexOptions.IgnoreCase);
            return reg.IsMatch(_data.Substring(_position));
        }
        public void Parse(string token)
        {
            SkipWhitespaces();
            if (EOF)
            {
                throw new EndOfCodeException();
            }
            if (!TryParse(token))
            {
                throw new InvalidTokenException(token, _line, _position, Resources.ERR_TOKENNM);
            }
            _position += token.Length;
            SkipWhitespaces();
        }
        public string GetNext(string rex)
        {
            if (EOF)
            {
                throw new EndOfCodeException();
            }
            if (!TryRegex(rex))
            {
                throw new NotMatchingCharacterException(_line, _position);
            }
            string nxt = _data.Substring(_position, 1);
            _position++;
            return nxt;
        }
        public void ThrowSyntaxException()
        {
            throw new CodeSyntaxErrorException(_line, _position);
        }
        public void SkipWhitespaces()
        {
            while ((_position < _data.Length) && char.IsWhiteSpace(_data, _position))
            {
                if (_data[_position] == '\n')
                {
                    _line++;
                }
                _position++;
            }
        }
    }
}
