﻿using System;
using System.Drawing;
using BNFUP.Exceptions;
using BNFUP.Interfaces;
using BNFUP.Properties;

namespace BNFUP.Rules
{
    [Serializable]
    public class AlternativeRules : RuleBase
    {
        public AlternativeRules() : base()
        {
        }
        public AlternativeRules(string name) : base(name)
        {
        }
        #region IRuleItem
        public override string Description
        {
            get
            {
                return string.Join(Resources.SEP_DESCRIPTOR, new string[] { Resources.DES_ALTRULE, Name });
            }
        }
        public override bool Multiple
        {
            get
            {
                return true;
            }
        }
        public override bool Matches(Parser parser)
        {
            foreach (IRuleItem ir in this)
            {
                if (ir.Matches(parser))
                {
                    return true;
                }
            }
            return false;
        }
        public override void Build(Parser parser)
        {
            if (Count > 0)
            {
                ICompilableObject obj = _object;
                base.Build(parser);  
                foreach (IRuleItem ir in this)
                {
                    parser.SavePosition();
                    try
                    {
                        if (ir.Matches(parser))
                        {
                            ir.Build(parser);
                            _text = parser.DiscardSavePoint();
                            CompleteObject();
                            return;
                        }
                    }
                    catch
                    {                        
                    }
                    parser.RestorePosition();
                }
                DiscardObject();
                _object = obj;
                throw new NoMatchingRuleException(Name);
            }
            else
            {
                throw new RuleUndefinedException(Name);
            }
        }
        #endregion
        #region IRuleDrawing
        public override SizeF MeasureRule(Graphics gr, Font fextra, Font fmain)
        {
            SizeF szr = base.MeasureRule(gr, fextra, fmain);
            SizeF szb = gr.MeasureString(" ", fextra);
            float w = 0f;
            float ws = gr.MeasureString("| ", fextra).Width;
            int celm = 0;
            foreach (IRuleItem ri in this)
            {
                IItemDrawing dw = ri as IItemDrawing;
                if (dw != null)
                {
                    SizeF sza = dw.Measure(gr, fextra, fmain);
                    if (sza.Width > w)
                    {
                        w = sza.Width;
                    }
                    celm++;
                }
            }
            float wo = 0f;
            return new SizeF(szr.Width + w + ws + wo, ((float)celm * szb.Height));
        }
        public override void DrawRule(Graphics gr, float x, float y, Font fextra, Font fmain)
        {
            SizeF szb = gr.MeasureString(" ", fextra);
            base.DrawRule(gr, x, y, fextra, fmain);
            x += base.MeasureRule(gr, fextra, fmain).Width;
            float xi = x;
            float xx = 0f;
            for (int ir = 0; ir < Count; ir++)
            {
                RuleBase rch = this[ir] as RuleBase;
                if (rch != null)
                {
                    if (ir > 0)
                    {
                        gr.DrawString("| ", fextra, Brushes.Black, xi, y);
                        xx = gr.MeasureString("| ", fextra).Width;
                    }
                    for (int ix = 0; ix < rch.Count; ix++)
                    {
                        IItemDrawing dw = rch[ix] as IItemDrawing;
                        if (dw != null)
                        {
                            dw.Draw(gr, x + xx, y, fextra, fmain);
                            x += dw.Measure(gr, fextra, fmain).Width;
                        }
                    }
                    x = xi;
                    if (ir < Count - 1)
                    {
                        y += rch.Measure(gr, fextra, fmain).Height;
                    }
                }
            }
        }
        #endregion
        #region IEditableItem
        public override bool CanCopy
        {
            get
            {
                return !(string.IsNullOrEmpty(Name) || (Name == Resources.NAME_DEFAULT) || (Count == 0));
            }
        }
        public override void AddChild(IRuleItem item)
        {
            if (item is RuleChain)
            {
                Add(item);
            }
        }
        public override void RemoveChild(int item)
        {
            RemoveAt(item);
        }
        public override void ReplaceChild(int item1, IRuleItem item2)
        {
            if (item2 is RuleChain)
            {
                this[item1] = item2;
            }
        }
        public override void InsertAfter(int item1, IRuleItem item2)
        {
            if (item2 is RuleChain)
            {
                Insert(item1, item2);
            }
        }
        public override bool ValidateType(Type t, bool generic)
        {
            return generic && (t == typeof(RuleChain));
        }
        #endregion
    }
}
