﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.ComponentModel;
using System.Windows.Forms;
using BNFUP.Exceptions;
using BNFUP.Interfaces;
using BNFUP.Attributes;
using BNFUP.Properties;

namespace BNFUP.Rules
{
    public delegate void IdChangedEventHandler(string newid);
    [Serializable]
    public abstract class ItemBase : IRuleItem, ICompilableObjectGenerator, IItemDrawing, IEditableItem
    {
        protected int _color = Color.Black.ToArgb();
        protected string _text = "";
        protected bool _completed = false;
        [NonSerialized]
        protected ICompilableObject _object = null;
        [NonSerialized]
        protected ICompilableObjectFactory _factory = null;
        public ItemBase()
        {
        }
        public event IdChangedEventHandler OnIdChanged = null;
        protected void ValidateID(string id)
        {
            OnIdChanged?.Invoke(id);
        }
        [Browsable(false)]
        public abstract string ID { get; }        
        #region IRuleItem
        [Browsable(false)]
        public bool Completed
        {
            get
            {
                return _completed;
            }
            set
            {
                _completed = value;
                if (!_completed)
                {
                    OnIdChanged = null;
                    OnCompleteObject = null;
                    OnCreateObject = null;
                    OnDiscardObject = null;
                }
            }
        }
        [CategoryGlobal("CAT_SYNTAX")]
        [DisplayNameGlobal("DIS_OPTIONAL")]
        [DescriptionGlobal("DES_OPTIONAL")]
        public bool Optional { get; set; }
        [Browsable(false)]
        public abstract string Description { get; }
        [Browsable(false)]
        public IEnumerable<IRuleItem> Childs
        {
            get
            {
                yield break;
            }
        }
        public string GetText()
        {
            return _text;
        }
        public void UpdateRule(string name, RuleBase rule, string current) { }
        public void ChangeRule(RuleBase oldrule, RuleBase newrule, string current) { }
        public void DeleteRule(RuleBase rule, string current) { }
        public abstract bool Matches(Parser parser);
        public virtual void Build(Parser parser)
        {
            if (_factory != null)
            {
                if (!string.IsNullOrEmpty(CompilableType))
                {
                    _object = _factory.CreateObject(CompilableType, this);
                    if (_object == null)
                    {
                        throw new UnknownObjectKeyException(CompilableType);
                    }
                    OnCreateObject?.Invoke(this);
                    OnCompleteObject?.Invoke(this);
                }
            }
            else
            {
                throw new ObjectFactoryNotDefinedException();
            }
        }
        #endregion
        #region ICompilableObjectGenerator
        [CategoryGlobal("CAT_COMPILER")]
        [DisplayNameGlobal("DIS_OBJTYPE")]
        [DescriptionGlobal("DES_OBJTYPE")]
        public string CompilableType { get; set; }
        [Browsable(false)]
        public ICompilableObjectFactory Factory
        {
            get
            {
                return _factory;
            }
            set
            {
                _factory = value;
            }
        }
        [Browsable(false)]
        public ICompilableObject Object
        {
            get
            {
                return _object;
            }
        }
        public void UpdateObject(ICompilableObject obj)
        {
            _object = obj;
        }
        [Browsable(false)]
        public event ObjectCreatedHandler OnCreateObject = null;
        [Browsable(false)]
        public event ObjectCreatedHandler OnDiscardObject = null;
        [Browsable(false)]
        public event ObjectCreatedHandler OnCompleteObject = null;
        #endregion
        #region IItemDrawing
        [CategoryGlobal("CAT_DISPLAY")]
        [DisplayNameGlobal("DIS_COLOR")]
        [DescriptionGlobal("DES_COLOR")]
        public Color ForeColor
        {
            get
            {
                return Color.FromArgb(_color);
            }
            set
            {
                _color = value.ToArgb();
            }
        }
        public abstract SizeF Measure(Graphics gr, Font fextra, Font fmain);
        public abstract void Draw(Graphics gr, float x, float y, Font fextra, Font fmain);
        #endregion
        #region IEditableItem
        [Browsable(false)]
        public abstract bool CanCopy { get; }
        [Browsable(false)]
        public int ItemIndex { get; set; }
        public IEnumerable<ToolStripItem> EditOptions(IEditableItem parent, IEditableItem gparent, IRuleItemClipboard clipboard)
        {
            ToolStripMenuItem ti = new ToolStripMenuItem(Resources.OPT_DELETE);
            ti.Tag = EditActions.Delete;
            yield return ti;            
            if (CanCopy && (clipboard != null))
            {
                yield return new ToolStripSeparator();
                ti = new ToolStripMenuItem(Resources.OPT_CUT);
                ti.Tag = EditActions.Cut;
                yield return ti;
                ti = new ToolStripMenuItem(Resources.OPT_COPY);
                ti.Tag = EditActions.Copy;
                yield return ti;
            }
            yield return new ToolStripSeparator();
            ti = new ToolStripMenuItem(Resources.OPT_ENLIST);
            ti.Tag = EditActions.Enlist;
            yield return ti;
            if (gparent != null)
            {
                ti = new ToolStripMenuItem(Resources.OPT_EXTRACT);
                ti.Tag = EditActions.Extract;
                yield return ti;
            }
            yield return new ToolStripSeparator();
            ti = new ToolStripMenuItem(Resources.OPT_MOVEUP);
            ti.Tag = EditActions.Up;
            yield return ti;
            ti = new ToolStripMenuItem(Resources.OPT_MOVEDOWN);
            ti.Tag = EditActions.Down;
            yield return ti;
            yield break;
        }
        public bool PerformAction(ToolStripItem action, IEditableItem parent, IEditableItem gparent, RuleTable rules, IRuleItemClipboard clipboard)
        {
            if (action.Tag is EditActions)
            {
                EditActions ca = (EditActions)action.Tag;
                switch (ca)
                {
                    case EditActions.Delete:
                        parent.RemoveChild(ItemIndex);
                        if (!CanCopy)
                        {
                            rules.RemoveItem(this);
                        }
                        break;
                    case EditActions.Cut:
                        parent.RemoveChild(ItemIndex);
                        clipboard.SetData(this);
                        break;
                    case EditActions.Copy:
                        clipboard.SetData(this);
                        break;
                    case EditActions.Enlist:
                        RuleItemList l = new RuleItemList();
                        l.Factory = _factory;
                        l.Add(this);
                        parent.ReplaceChild(ItemIndex, l);
                        break;
                    case EditActions.Extract:
                        if (gparent != null)
                        {
                            gparent.InsertAfter(parent.ItemIndex, this);
                            parent.RemoveChild(ItemIndex);
                        }
                        break;
                    case EditActions.Up:
                        parent.MoveUp(ItemIndex);
                        break;
                    case EditActions.Down:
                        parent.MoveDown(ItemIndex);
                        break;
                }
            }
            return false;
        }
        public void AddChild(IRuleItem item)
        {
            throw new NotImplementedException();
        }
        public void RemoveChild(int item)
        {
            throw new NotImplementedException();
        }
        public void ReplaceChild(int item1, IRuleItem item2)
        {
            throw new NotImplementedException();
        }
        public void InsertAfter(int item1, IRuleItem item2)
        {
            throw new NotImplementedException();
        }
        public void MoveUp(int item)
        {
            throw new NotImplementedException();
        }
        public void MoveDown(int item)
        {
            throw new NotImplementedException();
        }
        public bool ValidateType(Type t, bool generic)
        {
            return false;
        }
        #endregion
    }
}
