﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.ComponentModel;
using System.Windows.Forms;
using BNFUP.Properties;
using BNFUP.Interfaces;
using BNFUP.Attributes;
using BNFUP.Exceptions;
using BNFUP.Controls;

namespace BNFUP.Rules
{
    public delegate void RootChangedHandler(RuleBase newroot);
    [Serializable]
    public abstract class RuleBase : List<IRuleItem>, IRuleItem, IItemDrawing, IRuleDrawing, ICompilableObjectGenerator, IEditableItem, IEquatable<RuleBase>, IComparable<RuleBase>
    {
        [NonSerialized]
        protected ICompilableObjectFactory _factory = null;
        [NonSerialized]
        protected ICompilableObject _object = null;
        protected int _color = Color.Black.ToArgb();
        protected bool _root = false;
        protected string _text;
        protected bool _completed = false;
        public RuleBase()
        {
            Name = Resources.NAME_DEFAULT;
        }
        public RuleBase(string name)
        {
            Name = name;
        }
        [Browsable(false)]
        public new int Capacity
        {
            get
            {
                return base.Capacity;
            }
            set
            {
                base.Capacity = value;
            }
        }
        [Browsable(false)]
        public new int Count
        {
            get
            {
                return base.Count;
            }
        }
        #region Events
        public event RootChangedHandler OnRootChanged = null;
        #endregion
        #region Public Properties
        [CategoryGlobal("CAT_SYNTAX")]
        [DisplayNameGlobal("DIS_NAME")]
        [DescriptionGlobal("DES_NAME")]
        public string Name { get; set; }
        [Browsable(false)]
        public virtual bool Multiple
        {
            get
            {
                return false;
            }
        }
        [CategoryGlobal("CAT_SYNTAX")]
        [DisplayNameGlobal("DIS_ROOT")]
        [DescriptionGlobalAttribute("DES_ROOT")]
        public bool Root
        {
            get
            {
                return _root;
            }
            set
            {
                if (OnRootChanged != null)
                {
                    _root = value;
                    OnRootChanged(this);
                }
            }
        }
        [Browsable(false)]
        public IEnumerable<RuleBase> ChildRules
        {
            get
            {
                foreach (IRuleItem ri in this)
                {
                    if (ri is RuleBase)
                    {
                        yield return ri as RuleBase;
                    }
                }
            }
        }
        #endregion
        #region Protected Methods
        protected void DiscardObject()
        {
            if (!string.IsNullOrEmpty(CompilableType))
            {
                OnDiscardObject?.Invoke(this);
            }
        }
        protected void CompleteObject()
        {
            if (!string.IsNullOrEmpty(CompilableType))
            {
                OnCompleteObject?.Invoke(this);
            }
        }
        #endregion
        #region IRuleItem
        [Browsable(false)]
        public IEnumerable<IRuleItem> Childs
        {
            get
            {
                foreach (IRuleItem ri in this)
                {
                    yield return ri;
                }
            }
        }
        [Browsable(false)]
        public bool Completed
        {
            get
            {
                return _completed;
            }
            set
            {
                _completed = value;
                if (!_completed)
                {
                    OnCreateObject = null;
                    OnCompleteObject = null;
                    OnDiscardObject = null;
                    OnRootChanged = null;
                }
            }
        }
        [Browsable(false)]
        public bool Optional
        {
            get
            {
                return false;
            }
            set
            {
            }
        }
        [Browsable(false)]
        public abstract string Description { get; }
        public string GetText()
        {
            return _text;
        }
        public void UpdateRule(string name, RuleBase rule, string current)
        {
            if (current.Contains("{" + Name + "}"))
            {
                return;
            }
            for (int ix = 0; ix < Count; ix++)
            {
                TempRule tr = this[ix] as TempRule;
                if (tr != null)
                {
                    if (string.Compare(name, tr.Name, false) == 0)
                    {
                        this[ix] = rule;
                    }
                }
                else
                {
                    this[ix].UpdateRule(name, rule, current + "{" + Name + "}");
                }
            }
        }
        public void ChangeRule(RuleBase oldrule, RuleBase newrule, string current)
        {
            if (current.Contains("{" + Name + "}"))
            {
                return;
            }
            for (int ix = 0; ix < Count; ix++)
            {
                if (this[ix] == oldrule)
                {
                    this[ix] = newrule;
                }
                else
                {
                    this[ix].ChangeRule(oldrule, newrule, current + "{" + Name + "}");
                }
            }
        }
        public void DeleteRule(RuleBase rule, string current)
        {
            if (current.Contains("{" + Name + "}"))
            {
                return;
            }
            for (int ix = Count - 1; ix >= 0; ix--)
            {
                if (this[ix] == rule)
                {
                    RemoveAt(ix);
                }
                else
                {
                    this[ix].DeleteRule(rule, current + "{" + Name + "}");
                }
            }
        }
        public abstract bool Matches(Parser parser);
        public virtual void Build(Parser parser)
        {
            if (_factory != null)
            {
                if (!string.IsNullOrEmpty(CompilableType))
                {
                    _object = Factory.CreateObject(CompilableType, this);
                    if (_object == null)
                    {
                        throw new UnknownObjectKeyException(CompilableType);
                    }                        
                    OnCreateObject?.Invoke(this);
                }
            }
            else
            {
                throw new ObjectFactoryNotDefinedException();
            }
        }
        #endregion
        #region ICompilableObjectGenerator
        [CategoryGlobal("CAT_COMPILER")]
        [DisplayNameGlobal("DIS_OBJTYPE")]
        [DescriptionGlobal("DES_OBJTYPE")]
        public string CompilableType { get; set; }
        [Browsable(false)]
        public ICompilableObjectFactory Factory
        {
            get
            {
                return _factory;
            }
            set
            {
                _factory = value;
                foreach (IRuleItem ri in this)
                {
                    if ((ri is ICompilableObjectGenerator) && (((ICompilableObjectGenerator)ri).Factory != value))
                    {
                        ((ICompilableObjectGenerator)ri).Factory = value;
                    }
                }
            }
        }
        [Browsable(false)]
        public ICompilableObject Object
        {
            get
            {
                return _object;
            }
        }
        public void UpdateObject(ICompilableObject obj)
        {
            _object = obj;
        }
        [Browsable(false)]
        public event ObjectCreatedHandler OnCreateObject = null;
        [Browsable(false)]
        public event ObjectCreatedHandler OnDiscardObject = null;
        [Browsable(false)]
        public event ObjectCreatedHandler OnCompleteObject = null;
        #endregion
        #region IItemDrawing
        [CategoryGlobal("CAT_DISPLAY")]
        [DisplayNameGlobal("DIS_COLOR")]
        [DescriptionGlobal("DES_COLOR")]
        public Color ForeColor
        {
            get
            {
                return Color.FromArgb(_color);
            }
            set
            {
                _color = value.ToArgb();
            }
        }
        public SizeF Measure(Graphics gr, Font fextra, Font fmain)
        {
            SizeF szi = gr.MeasureString("<", fextra);
            SizeF sz = gr.MeasureString(Name, fmain);
            SizeF szf = gr.MeasureString(">", fextra);
            return new SizeF(szi.Width + sz.Width + szf.Width, sz.Height);
        }
        public void Draw(Graphics gr, float x, float y, Font fextra, Font fmain)
        {
            SizeF szi = gr.MeasureString("<", fextra);
            SizeF szf = gr.MeasureString(">", fextra);
            SizeF sz = gr.MeasureString(Name, fmain);
            Brush bc = new SolidBrush(this is TempRule ? Color.Red : ForeColor);
            gr.DrawString("<", fextra, Brushes.Black, x, y);
            x += szi.Width;
            gr.DrawString(Name, fmain, bc, x, y);
            x += sz.Width;
            gr.DrawString(">", fextra, Brushes.Black, x, y);
            x += szf.Width;
            bc.Dispose();
        }
        #endregion
        #region IRuleDrawing
        public virtual SizeF MeasureRule(Graphics gr, Font fextra, Font fmain)
        {
            SizeF szi = gr.MeasureString(Root ? "<<" : "<", fextra);
            SizeF sz = gr.MeasureString(Name, fmain);
            SizeF szf = gr.MeasureString(Root ? ">>" : ">", fextra);
            SizeF szb = gr.MeasureString(" ", fextra);
            SizeF sze = gr.MeasureString("::=", fextra);
            return new SizeF(szi.Width + sz.Width + szf.Width + szb.Width + sze.Width + szb.Width, szb.Height);
        }
        public virtual void DrawRule(Graphics gr, float x, float y, Font fextra, Font fmain)
        {
            SizeF szi = gr.MeasureString(Root ? "<<" : "<", fextra);
            SizeF szf = gr.MeasureString(Root ? ">>" : ">", fextra);
            SizeF sz = gr.MeasureString(Name, fmain);
            SizeF szb = gr.MeasureString(" ", fextra);
            SizeF sze = gr.MeasureString("::=", fextra);
            Brush bc = new SolidBrush(ForeColor);
            gr.DrawString(Root ? "<<" : "<", fextra, Brushes.Black, x, y);
            x += szi.Width;
            gr.DrawString(Name, fmain, bc, x, y);
            x += sz.Width;
            gr.DrawString(Root ? ">>" : ">", fextra, Brushes.Black, x, y);
            x += szf.Width + szb.Width;
            gr.DrawString("::=", fextra, Brushes.Black, x, y);
            bc.Dispose();
        }
        #endregion
        #region IEditableItem
        [Browsable(false)]
        public abstract bool CanCopy { get; }
        [Browsable(false)]
        public int ItemIndex { get; set; }
        public IEnumerable<ToolStripItem> EditOptions(IEditableItem parent, IEditableItem gparent, IRuleItemClipboard clipboard)
        {
            ToolStripMenuItem ti = new ToolStripMenuItem(Resources.OPT_NEW);
            ti.Tag = EditActions.New;
            yield return ti;
            if (parent != null)
            {
                ti = new ToolStripMenuItem(Resources.OPT_DELETE);
                ti.Tag = EditActions.Delete;
                yield return ti;
            }
            bool cc = CanCopy;
            if (cc && (clipboard != null))
            {
                yield return new ToolStripSeparator();
                if (parent != null)
                {                    
                    ti = new ToolStripMenuItem(Resources.OPT_CUT);
                    ti.Tag = EditActions.Cut;
                    yield return ti;
                }
                ti = new ToolStripMenuItem(Resources.OPT_COPY);
                ti.Tag = EditActions.Copy;
                yield return ti;
            }
            if ((clipboard != null) && clipboard.ContainsData)
            {
                IRuleItem ir = clipboard.GetData();
                if (ir != null)
                {
                    if ((!Multiple) || (ir is RuleChain))
                    {
                        if (!cc)
                        {
                            yield return new ToolStripSeparator();
                        }
                        ti = new ToolStripMenuItem(Resources.OPT_PASTE);
                        ti.Tag = EditActions.Paste;
                        yield return ti;
                    }
                }
            }
            if (parent != null)
            {
                if (!((parent is RuleBase) && (((RuleBase)parent).Multiple)))
                {
                    yield return new ToolStripSeparator();
                    ti = new ToolStripMenuItem(Resources.OPT_ENLIST);
                    ti.Tag = EditActions.Enlist;
                    yield return ti;
                    if (gparent != null)
                    {
                        ti = new ToolStripMenuItem(Resources.OPT_EXTRACT);
                        ti.Tag = EditActions.Extract;
                        yield return ti;
                    }
                }
                yield return new ToolStripSeparator();
                ti = new ToolStripMenuItem(Resources.OPT_MOVEUP);
                ti.Tag = EditActions.Up;
                yield return ti;
                ti = new ToolStripMenuItem(Resources.OPT_MOVEDOWN);
                ti.Tag = EditActions.Down;
                yield return ti;
            }
            else
            {
                if ((Multiple) && (Count == 1))
                {
                    yield return new ToolStripSeparator();
                    ti = new ToolStripMenuItem(Resources.OPT_SIMPLIFY);
                    ti.Tag = EditActions.Simplify;
                    yield return ti;
                }
                if (!Multiple)
                {
                    yield return new ToolStripSeparator();
                    ti = new ToolStripMenuItem(Resources.OPT_ADDOPTS);
                    ti.Tag = EditActions.Alternatives;
                    yield return ti;
                }
            }
            yield break;
        }
        public bool PerformAction(ToolStripItem action, IEditableItem parent, IEditableItem gparent, RuleTable rules, IRuleItemClipboard clipboard)
        {
            if (action.Tag is EditActions)
            {
                EditActions ca = (EditActions)action.Tag;
                switch (ca)
                {
                    case EditActions.Delete:
                        parent.RemoveChild(ItemIndex);
                        break;
                    case EditActions.Cut:
                        parent.RemoveChild(ItemIndex);
                        clipboard.SetData(this);
                        break;
                    case EditActions.Copy:
                        clipboard.SetData(this);
                        break;
                    case EditActions.Paste:
                        if (clipboard.ContainsData)
                        {
                            IRuleItem ir = clipboard.GetData();
                            if (ir is ICompilableObjectGenerator)
                            {
                                ((ICompilableObjectGenerator)ir).Factory = _factory;
                            }
                            if (ir is ItemBase)
                            {
                                AddChild(rules.RestoreItem(ir));
                            }
                            else if (ir is TokenList)
                            {
                                ir.Completed = false;
                                TokenList tl = rules.RestoreItem(ir) as TokenList;
                                rules.LinkItem(tl);
                                AddChild(tl);
                            }
                            else if (ir is RuleBase)
                            {
                                AddChild(rules.RestoreItem(ir));
                            }
                            else if (ir is RuleItemList)
                            {
                                RuleItemList rl = rules.RestoreItem(ir) as RuleItemList;
                                rl.Completed = false;
                                AddChild(rl);
                                rules.LinkItem(rl);
                            }
                        }
                        break;
                    case EditActions.Enlist:
                        RuleItemList l = new RuleItemList();
                        l.Factory = _factory;
                        l.Add(this);
                        parent.ReplaceChild(ItemIndex, l);
                        break;
                    case EditActions.Extract:
                        if (gparent != null)
                        {
                            gparent.InsertAfter(parent.ItemIndex, this);
                            parent.RemoveChild(ItemIndex);
                        }
                        break;
                    case EditActions.Simplify:
                        RuleBase rbc = this[0] as RuleBase;
                        if (rbc != null)
                        {
                            rbc.Name = Name;
                            rbc.CompilableType = CompilableType;
                            rbc.Root = Root;
                            rules.RefreshRule(rbc, this);
                        }
                        return true;
                    case EditActions.Alternatives:
                        AlternativeRules ar = new AlternativeRules(Name);
                        ar.Factory = _factory;                   
                        ar.Root = Root;
                        ar.CompilableType = CompilableType;
                        CompilableType = "";
                        rules.RefreshRule(ar, this);
                        Name = Name + "1";
                        ar.AddChild(this);
                        Completed = false;
                        rules.LinkItem(this);
                        return true;
                    case EditActions.Up:
                        parent.MoveUp(ItemIndex);
                        break;
                    case EditActions.Down:
                        parent.MoveDown(ItemIndex);
                        break;
                    case EditActions.New:
                        dlgItems dlg = new dlgItems();
                        dlg.ListOwner = this;
                        dlg.Rules = rules;
                        dlg.InsertTop = Count;
                        if (dlg.ShowDialog() == DialogResult.OK)
                        {
                            IRuleItem ir = dlg.SelectedObject;
                            if (ir is ICompilableObjectGenerator)
                            {
                                ((ICompilableObjectGenerator)ir).Factory = _factory;
                            }
                            if (ir != null)
                            {
                                InsertAfter(dlg.InsertPos, ir);
                                if (ir is ItemBase)
                                {
                                    rules.AddItem(ir);
                                }
                                else
                                {
                                    rules.LinkItem(ir);
                                }
                            }
                        }
                        break;
                }
            }
            return false;
        }
        public abstract void AddChild(IRuleItem item);
        public abstract void RemoveChild(int item);
        public abstract void ReplaceChild(int item1, IRuleItem item2);
        public abstract void InsertAfter(int item1, IRuleItem item2);
        public void MoveUp(int item)
        {
            if (item > 0)
            {
                IRuleItem tmp = this[item - 1];
                this[item - 1] = this[item];
                this[item] = tmp;
            }
        }
        public void MoveDown(int item)
        {
            if (item < Count - 1)
            {
                IRuleItem tmp = this[item + 1];
                this[item + 1] = this[item];
                this[item] = tmp;
            }
        }
        public abstract bool ValidateType(Type t, bool generic);
        #endregion
        #region IEquatable
        public bool Equals(RuleBase other)
        {
            return Name == other.Name;
        }
        #endregion
        #region IComparable
        public int CompareTo(RuleBase other)
        {
            return string.Compare(Name, other.Name, true);
        }
        #endregion
    }    
}
