﻿using System;
using System.Drawing;
using BNFUP.Exceptions;
using BNFUP.Interfaces;
using BNFUP.Properties;

namespace BNFUP.Rules
{
    [Serializable]
    public class RuleChain : RuleBase
    {
        public RuleChain() : base()
        {
        }
        public RuleChain(string name) : base(name)
        {
        }
        #region IRuleItem
        public override string Description
        {
            get
            {
                return string.Join(Resources.SEP_DESCRIPTOR, new string[] { Resources.DES_RULE, Name });
            }
        }
        public override bool Matches(Parser parser)
        {
            foreach (IRuleItem ri in this)
            {
                if (ri.Optional)
                {
                    if (ri.Matches(parser))
                    {
                        return true;
                    }
                }
                else
                {
                    return ri.Matches(parser);
                }
            }
            return false;
        }
        public override void Build(Parser parser)
        {
            if (Count > 0)
            {
                ICompilableObject obj = _object;
                base.Build(parser);
                parser.SavePosition();                
                foreach (IRuleItem ri in this)
                {
                    try
                    {
                        ri.Build(parser);
                    }
                    catch
                    {
                        parser.DiscardSavePoint();
                        DiscardObject();
                        _object = obj;
                        throw;
                    }
                }
                _text = parser.DiscardSavePoint();
                CompleteObject();
            }
            else
            {
                throw new RuleUndefinedException(Name);
            }
        }
        #endregion
        #region IRuleDrawing
        public override SizeF MeasureRule(Graphics gr, Font fextra, Font fmain)
        {
            SizeF szr = base.MeasureRule(gr, fextra, fmain);
            SizeF szb = gr.MeasureString(" ", fextra);
            float w = szr.Width;
            float h = szb.Height;
            int celm = 0;
            foreach (IRuleItem ri in this)
            {
                IItemDrawing dw = ri as IItemDrawing;
                if (dw != null)
                {
                    w += dw.Measure(gr, fextra, fmain).Width;
                    celm++;
                }
            }
            return new SizeF(w + ((float)(celm - 1) * szb.Width), h);
        }
        public override void DrawRule(Graphics gr, float x, float y, Font fextra, Font fmain)
        {
            SizeF szb = gr.MeasureString(" ", fextra);
            base.DrawRule(gr, x, y, fextra, fmain);
            x += base.MeasureRule(gr, fextra, fmain).Width;
            for (int ix = 0; ix < Count; ix++)
            {
                IItemDrawing dw = this[ix] as IItemDrawing;
                if (dw != null)
                {
                    dw.Draw(gr, x, y, fextra, fmain);
                    x += dw.Measure(gr, fextra, fmain).Width;
                    if (ix < Count - 1)
                    {
                        x += szb.Width;
                    }
                }
            }
        }
        #endregion
        #region IEditableItem
        public override bool CanCopy
        {
            get
            {
                return !(string.IsNullOrEmpty(Name) && (Name != Resources.NAME_DEFAULT));
            }
        }
        public override void AddChild(IRuleItem item)
        {
            Add(item);
        }
        public override void RemoveChild(int item)
        {
            RemoveAt(item);
        }
        public override void ReplaceChild(int item1, IRuleItem item2)
        {
            this[item1] = item2;
        }
        public override void InsertAfter(int item1, IRuleItem item2)
        {
            Insert(item1, item2);
        }
        public override bool ValidateType(Type t, bool generic)
        {
            return !generic || !t.IsSubclassOf(typeof(RuleBase));
        }
        #endregion
    }
}
