﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.ComponentModel;
using System.Windows.Forms;
using BNFUP.Interfaces;
using BNFUP.Exceptions;
using BNFUP.Properties;
using BNFUP.Attributes;
using BNFUP.Controls;

namespace BNFUP.Rules
{
    [Serializable]
    public class RuleItemList : SimpleRuleItemList, IRuleItem, IItemDrawing , ICompilableObjectGenerator, IEditableItem
    {
        private int _color = Color.Black.ToArgb();
        private bool _completed = false;
        [NonSerialized]
        private ICompilableObject _object = null;
        [NonSerialized]
        private ICompilableObjectFactory _factory = null;
        private string _text;

        public RuleItemList()
        {
        }
        #region IRuleItem
        [Browsable(false)]
        public bool Completed
        {
            get
            {
                return _completed;
            }
            set
            {
                _completed = value;
                if (!_completed)
                {
                    OnCompleteObject = null;
                    OnCreateObject = null;
                    OnDiscardObject = null;
                    foreach (IRuleItem ir in this)
                    {
                        if ((ir is RuleItemList) || (ir is TokenList))
                        {
                            ir.Completed = false;
                        }
                    }
                }
            }
        }
        [CategoryGlobal("CAT_SYNTAX")]
        [DisplayNameGlobal("DIS_OPTIONAL")]
        [DescriptionGlobal("DES_OPTIONAL")]
        public bool Optional { get; set; }
        [Browsable(false)]
        public string Description
        {
            get
            {
                return Resources.DES_ITEMLIST;
            }
        }
        [Browsable(false)]
        public IEnumerable<IRuleItem> Childs
        {
            get
            {
                foreach (IRuleItem ri in this)
                {
                    yield return ri;
                }
            }
        }
        public string GetText()
        {
            return _text;
        }
        public void UpdateRule(string name, RuleBase rule, string current)
        {
            for (int ix = 0; ix < Count; ix++)
            {
                TempRule tr = this[ix] as TempRule;
                if (tr != null)
                {
                    if (string.Compare(name, tr.Name, false) == 0)
                    {
                        this[ix] = rule;
                    }
                }
                else
                {
                    this[ix].UpdateRule(name, rule, current);
                }
            }
        }
        public void ChangeRule(RuleBase oldrule, RuleBase newrule, string current)
        {
            for (int ix = 0; ix < Count; ix++)
            {
                if (this[ix] == oldrule)
                {
                    this[ix] = newrule;
                }
                else
                {
                    this[ix].ChangeRule(oldrule, newrule, current);
                }
            }
        }
        public void DeleteRule(RuleBase rule, string current)
        {
            for (int ix = Count - 1; ix >= 0; ix--)
            {
                if (this[ix] == rule)
                {
                    RemoveAt(ix);
                }
                else
                {
                    this[ix].DeleteRule(rule, current);
                }
            }
        }
        public bool Matches(Parser parser)
        {
            if (Count > 0)
            {
                for (int ix = 0; ix < Count; ix++)
                {
                    if (this[ix].Matches(parser))
                    {
                        return true;
                    }
                    if (!this[ix].Optional)
                    {
                        break;
                    }
                }
            }
            return false;
        }
        public void Build(Parser parser)
        {
            ICompilableObject obj = _object;
            if (_factory != null)
            {
                if (!string.IsNullOrEmpty(CompilableType))
                {
                    _object = _factory.CreateObject(CompilableType, this);
                    if (_object == null)
                    {
                        throw new UnknownObjectKeyException(CompilableType);
                    }
                    OnCreateObject?.Invoke(this);
                }
            }
            else
            {
                throw new ObjectFactoryNotDefinedException();
            }
            parser.SavePosition();
            foreach (IRuleItem ir in this)
            {
                try
                {
                    ir.Build(parser);
                }
                catch
                {
                    if (!ir.Optional)
                    {
                        if (Optional)
                        {
                            parser.RestorePosition();
                            if (!string.IsNullOrEmpty(CompilableType))
                            {
                                OnDiscardObject?.Invoke(this);
                                _object = obj;
                            }
                            return;
                        }
                        else
                        {
                            parser.DiscardSavePoint();
                            if (!string.IsNullOrEmpty(CompilableType))
                            {
                                OnDiscardObject?.Invoke(this);
                                _object = obj;
                            }
                            throw;
                        }
                    }
                }
            }
            _text = parser.DiscardSavePoint();
            if (!string.IsNullOrEmpty(CompilableType))
            {
                OnCompleteObject?.Invoke(this);
            }
        }
        #endregion
        #region ICompilableObjectGenerator
        [CategoryGlobal("CAT_COMPILER")]
        [DisplayNameGlobal("DIS_OBJTYPE")]
        [DescriptionGlobal("DES_OBJTYPE")]
        public string CompilableType { get; set; }
        [Browsable(false)]
        public ICompilableObjectFactory Factory
        {
            get
            {
                return _factory;
            }
            set
            {
                _factory = value;
                foreach (IRuleItem ri in this)
                {
                    if ((ri is ICompilableObjectGenerator) && (((ICompilableObjectGenerator)ri).Factory != value))
                    {
                        ((ICompilableObjectGenerator)ri).Factory = value;
                    }
                }
            }
        }
        [Browsable(false)]
        public ICompilableObject Object
        {
            get
            {
                return _object;
            }
        }
        public void UpdateObject(ICompilableObject obj)
        {
            _object = obj;
        }
        [Browsable(false)]
        public event ObjectCreatedHandler OnCreateObject = null;
        [Browsable(false)]
        public event ObjectCreatedHandler OnDiscardObject = null;
        [Browsable(false)]
        public event ObjectCreatedHandler OnCompleteObject = null;
        #endregion
        #region IItemDrawing
        [CategoryGlobal("CAT_DISPLAY")]
        [DisplayNameGlobal("DIS_COLOR")]
        [DescriptionGlobal("DES_COLOR")]
        public Color ForeColor
        {
            get
            {
                return Color.FromArgb(_color);
            }
            set
            {
                _color = value.ToArgb();
            }
        }
        public SizeF Measure(Graphics gr, Font fextra, Font fmain)
        {
            SizeF szb = gr.MeasureString(" ", fextra);
            float w = 0f;
            float h = szb.Height;
            int celm = 0;            
            foreach (IRuleItem ri in this)
            {
                IItemDrawing dw = ri as IItemDrawing;
                if (dw != null)
                {
                    w += dw.Measure(gr, fextra, fmain).Width;
                    celm++;
                }
            }
            float wo = 0f;
            if (Optional)
            {
                wo = gr.MeasureString("[ ", fextra).Width + gr.MeasureString(" ]", fextra).Width;
            }
            return new SizeF(w + wo + ((float)(celm - 1) * szb.Width), h);
        }
        public void Draw(Graphics gr, float x, float y, Font fextra, Font fmain)
        {
            SizeF szb = gr.MeasureString(" ", fextra);
            if (Optional)
            {
                gr.DrawString("[ ", fextra, Brushes.Black, x, y);
                x += gr.MeasureString("[ ", fextra).Width;
            }
            for (int ix = 0; ix < Count; ix++)
            {
                IItemDrawing dw = this[ix] as IItemDrawing;
                if (dw != null)
                {
                    dw.Draw(gr, x, y, fextra, fmain);
                    x += dw.Measure(gr, fextra, fmain).Width;
                    if (ix < Count - 1)
                    {
                        x += szb.Width;                        
                    }
                }
            }
            if (Optional)
            {
                gr.DrawString(" ]", fextra, Brushes.Black, x, y);
            }
        }
        #endregion
        #region IEditableItem
        [Browsable(false)]
        public bool CanCopy
        {
            get
            {
                foreach (IRuleItem ir in this)
                {
                    IEditableItem ei = ir as IEditableItem;
                    if ((ei != null) && (!ei.CanCopy))
                    {
                        return false;
                    }
                }
                return true;
            }
        }
        [Browsable(false)]
        public int ItemIndex { get; set; }
        public IEnumerable<ToolStripItem> EditOptions(IEditableItem parent, IEditableItem gparent, IRuleItemClipboard clipboard)
        {
            ToolStripMenuItem ti = new ToolStripMenuItem(Resources.OPT_NEW);
            ti.Tag = EditActions.New;
            yield return ti;
            ti = new ToolStripMenuItem(Resources.OPT_DELETE);
            ti.Tag = EditActions.Delete;
            yield return ti;
            ti = new ToolStripMenuItem(Resources.OPT_FREE);
            ti.Tag = EditActions.Free;
            yield return ti;
            bool cc = CanCopy;
            if (cc && (clipboard != null))
            {
                yield return new ToolStripSeparator();
                ti = new ToolStripMenuItem(Resources.OPT_CUT);
                ti.Tag = EditActions.Cut;
                yield return ti;
                ti = new ToolStripMenuItem(Resources.OPT_COPY);
                ti.Tag = EditActions.Copy;
                yield return ti;
            }
            if ((clipboard != null) && clipboard.ContainsData)
            {
                IRuleItem ir = clipboard.GetData();
                if (ir != null)
                {
                    if (!cc)
                    {
                        yield return new ToolStripSeparator();
                    }
                    ti = new ToolStripMenuItem(Resources.OPT_PASTE);
                    ti.Tag = EditActions.Paste;
                    yield return ti;
                }
            }
            yield return new ToolStripSeparator();
            ti = new ToolStripMenuItem(Resources.OPT_ENLIST);
            ti.Tag = EditActions.Enlist;
            yield return ti;
            if (gparent != null)
            {
                ti = new ToolStripMenuItem(Resources.OPT_EXTRACT);
                ti.Tag = EditActions.Extract;
                yield return ti;
            }
            yield return new ToolStripSeparator();
            ti = new ToolStripMenuItem(Resources.OPT_MOVEUP);
            ti.Tag = EditActions.Up;
            yield return ti;
            ti = new ToolStripMenuItem(Resources.OPT_MOVEDOWN);
            ti.Tag = EditActions.Down;
            yield return ti;
            yield break;
        }
        public bool PerformAction(ToolStripItem action, IEditableItem parent, IEditableItem gparent, RuleTable rules, IRuleItemClipboard clipboard)
        {
            if (action.Tag is EditActions)
            {
                EditActions ca = (EditActions)action.Tag;
                switch (ca)
                {
                    case EditActions.Delete:
                        parent.RemoveChild(ItemIndex);
                        break;
                    case EditActions.Cut:
                        parent.RemoveChild(ItemIndex);
                        clipboard.SetData(this);
                        break;
                    case EditActions.Copy:
                        clipboard.SetData(this);
                        break;
                    case EditActions.Paste:
                        if (clipboard.ContainsData)
                        {
                            IRuleItem ir = clipboard.GetData();
                            if (ir is ICompilableObjectGenerator)
                            {
                                ((ICompilableObjectGenerator)ir).Factory = _factory;
                            }
                            if (ir is ItemBase)
                            {
                                AddChild(rules.RestoreItem(ir));
                            }
                            else if (ir is TokenList)
                            {
                                ir.Completed = false;
                                TokenList tl = rules.RestoreItem(ir) as TokenList;
                                rules.LinkItem(tl);
                                AddChild(tl);
                            }
                            else if (ir is RuleBase)
                            {
                                RuleBase rb = rules.RestoreItem(ir) as RuleBase;
                                AddChild(rb);
                            }
                            else if (ir is RuleItemList)
                            {
                                RuleItemList rl = rules.RestoreItem(ir) as RuleItemList;
                                rl.Completed = false;
                                AddChild(rl);
                                rules.LinkItem(rl);
                            }
                        }
                        break;
                    case EditActions.Enlist:
                        RuleItemList l = new RuleItemList();
                        l.Factory = _factory;
                        l.Add(this);
                        parent.ReplaceChild(ItemIndex, l);
                        break;
                    case EditActions.Extract:
                        if (gparent != null)
                        {
                            gparent.InsertAfter(parent.ItemIndex, this);
                            parent.RemoveChild(ItemIndex);
                        }
                        break;
                    case EditActions.Free:
                        int ix = ItemIndex + 1;
                        foreach (IRuleItem ri in this)
                        {
                            parent.InsertAfter(ix++, ri);
                        }
                        parent.RemoveChild(ItemIndex);
                        break;
                    case EditActions.Up:
                        parent.MoveUp(ItemIndex);
                        break;
                    case EditActions.Down:
                        parent.MoveDown(ItemIndex);
                        break;
                    case EditActions.New:
                        dlgItems dlg = new dlgItems();
                        dlg.ListOwner = this;
                        dlg.Rules = rules;
                        dlg.InsertTop = Count;
                        if (dlg.ShowDialog() == DialogResult.OK)
                        {
                            IRuleItem ir = dlg.SelectedObject;
                            if (ir is ICompilableObjectGenerator)
                            {
                                ((ICompilableObjectGenerator)ir).Factory = _factory;
                            }
                            if (ir != null)
                            {
                                InsertAfter(dlg.InsertPos, ir);
                                if (ir is ItemBase)
                                {
                                    rules.AddItem(ir);
                                }
                                else
                                {
                                    rules.LinkItem(ir);
                                }
                            }
                        }
                        break;
                }
            }
            return false;
        }
        public void AddChild(IRuleItem item)
        {
            Add(item);
        }
        public void RemoveChild(int item)
        {
            RemoveAt(item);
        }
        public void ReplaceChild(int item1, IRuleItem item2)
        {
            this[item1] = item2;
        }
        public void InsertAfter(int item1, IRuleItem item2)
        {
            Insert(item1, item2);
        }
        public void MoveUp(int item)
        {
            if (item > 0)
            {
                IRuleItem tmp = this[item - 1];
                this[item - 1] = this[item];
                this[item] = tmp;
            }
        }
        public void MoveDown(int item)
        {
            if (item < Count - 1)
            {
                IRuleItem tmp = this[item + 1];
                this[item + 1] = this[item];
                this[item] = tmp;
            }
        }
        public bool ValidateType(Type t, bool generic)
        {
            return !generic || !t.IsSubclassOf(typeof(RuleBase));
        }
        #endregion
    }
}
