﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Runtime.Serialization.Formatters.Binary;
using BNFUP.Exceptions;
using BNFUP.Interfaces;
using BNFUP.Properties;

namespace BNFUP.Rules
{
    public delegate void CompilerEventHandler(string ev);
    internal class CompilerObject
    {
        public CompilerObject(ICompilableObjectGenerator owner, ICompilableObject obj)
        {
            Owner = owner;
            Object = obj;
        }
        public ICompilableObjectGenerator Owner { get; set; }
        public ICompilableObject Object { get; set; }
    }
    [Serializable]
    public class RuleTable
    {
        // List of rules
        private List<RuleBase> _rules = new List<RuleBase>();
        // List of tokens and character sets
        private SimpleRuleItemList _items = new SimpleRuleItemList();
        // Root rule, where the parser starts 
        private RuleBase _root = null;
        private string _rootname;
        private bool _caseSensitive = false;
        [NonSerialized]
        private ICompilableObjectFactory _factory = null;
        [NonSerialized]
        private CompilerObject _current = null;
        // This stack stores the objects that the compiler goes creating
        [NonSerialized]
        private Stack<CompilerObject> _stack = null;

        public RuleTable()
        {            
        }
        public RuleTable(ICompilableObjectFactory factory)
        {
            _factory = factory;
        }
        /// <summary>
        /// With this constructor, you can create the rules from a text file 
        /// </summary>
        /// <param name="rdr">TextReader to access the BNF source code</param>
        /// <param name="factory">ICompilableObject generator</param>
        public RuleTable(TextReader rdr, ICompilableObjectFactory factory)
        {
            _factory = factory;
            Parser par = new Parser(rdr) { CaseSensitive = _caseSensitive };
            RRules(par);
            if (!string.IsNullOrEmpty(_rootname))
            {
                RuleBase rb = GetRule(_rootname);
                if (rb != null)
                {
                    rb.Root = true;
                }
            }
            if (_root == null)
            {
                throw new RootRuleNotDefinedException();
            }
            foreach (RuleBase rb in _rules)
            {
                foreach (RuleBase rn in _rules)
                {
                    rb.UpdateRule(rn.Name, rn, "");
                }
            }
        }
        /// <summary>
        /// Create a RuleTable from a .bnf file
        /// </summary>
        /// <param name="bnffile">bnf file path</param>
        /// <returns>RuleTable object contained in the file</returns>
        public static RuleTable FromFile(string bnffile)
        {
            FileStream f = null;
            try
            {
                f = new FileStream(bnffile, FileMode.Open, FileAccess.Read, FileShare.Read);
                BinaryFormatter fmt = new BinaryFormatter();
                RuleTable rt = (RuleTable)fmt.Deserialize(f);
                return rt;
            }
            finally
            {
                if (f != null)
                {
                    f.Close();
                }
            }
        }
        /// <summary>
        /// Create a RuleTable from a .bnf file in a Stream
        /// </summary>
        /// <param name="bnffile">bnf file Stream</param>
        /// <returns>RuleTable object contained in the Stream</returns>
        public static RuleTable FromStream(Stream bnffile)
        {
            try
            {
                BinaryFormatter fmt = new BinaryFormatter();
                RuleTable rt = (RuleTable)fmt.Deserialize(bnffile);
                return rt;
            }
            finally
            {                
                bnffile.Close();
            }
        }
        /// <summary>
        /// Use this event to get compilation messages
        /// </summary>
        public event CompilerEventHandler OnCompilationFeedback = null;
        #region Public Properties
        public bool CaseSensitive
        {
            get
            {
                return _caseSensitive;
            }
            set
            {
                _caseSensitive = value;
            }
        }
        /// <summary>
        /// Gets or deactivates the debug mode
        /// </summary>
        public bool DebugMode
        {
            get
            {
                return OnCompilationFeedback != null;
            }
            set
            {
                if (!value)
                {
                    OnCompilationFeedback = null;
                }
            }
        }
        /// <summary>
        /// get or set the object generator 
        /// </summary>
        public ICompilableObjectFactory Factory
        {
            get
            {
                return _factory;
            }
            set
            {
                _factory = value;
                foreach (RuleBase rb in _rules)
                {
                    rb.Factory = value;
                }
            }
        }
        /// <summary>
        /// Get the Root rule, where the parsing starts
        /// </summary>
        public RuleBase Root
        {
            get
            {
                return _root;
            }
        }
        /// <summary>
        /// List all the Types of objects implementing the IRuleItem interface
        /// </summary>
        public IEnumerable<Type> ItemTypes
        {
            get
            {
                foreach (Type t in Assembly.GetExecutingAssembly().GetTypes())
                {
                    if (!t.IsAbstract && (t.GetInterface("IRuleItem") != null) && (t != typeof(TempRule)))
                    {
                        yield return t;
                    }
                }
            }
        }
        /// <summary>
        /// Enumerate the simple objects: tokens and character sets
        /// </summary>
        public IEnumerable<IRuleItem> SimpleItems
        {
            get
            {
                foreach (IRuleItem ri in _items)
                {
                    yield return ri;
                }
            }
        }
        /// <summary>
        /// Enumerate all the rules in the table
        /// </summary>
        public IEnumerable<RuleBase> Rules
        {
            get
            {
                foreach (RuleBase rb in _rules)
                {
                    yield return rb;
                }
            }
        }
        #endregion
        #region Public Methods
        /// <summary>
        /// Add a new rule to the table
        /// </summary>
        /// <param name="rule">New rule to be added</param>
        public void AddRule(RuleBase rule)
        {
            AddRule(rule, _rules);
        }
        /// <summary>
        /// Change the definition of a rule
        /// </summary>
        /// <param name="newrule">The new version of the rule</param>
        /// <param name="oldrule">The old version of the rule</param>
        public void RefreshRule(RuleBase newrule, RuleBase oldrule)
        {
            RefreshRule(newrule, oldrule, _rules);
        }
        /// <summary>
        /// Add a new simple item: token or character set
        /// </summary>
        /// <param name="item">New item to be added</param>
        public void AddItem(IRuleItem item)
        {
            if (!_items.Contains(item))
            {
                LinkItem(item);
                _items.Add(item);
            }
        }
        /// <summary>
        /// Remove one rule from the table and every of his references
        /// </summary>
        /// <param name="rule">Rule to remove</param>
        public void RemoveRule(RuleBase rule)
        {
            if (_rules.Contains(rule))
            {
                RuleBase rr = _rules[_rules.IndexOf(rule)];
                _rules.RemoveAt(_rules.IndexOf(rule));
                foreach (RuleBase r in _rules)
                {
                    r.DeleteRule(rr, "");
                }
            }
        }
        /// <summary>
        /// Remove one item from the item list
        /// </summary>
        /// <param name="item">Item to remove</param>
        public void RemoveItem(ItemBase item)
        {
            for (int ix = 0; ix < _items.Count; ix++)
            {
                if (_items[ix] == item)
                {
                    _items.RemoveAt(ix);
                    return;
                }
            }
        }
        /// <summary>
        /// Get a rule by his name
        /// </summary>
        /// <param name="name">Name of the rule</param>
        /// <returns>The rule or null if no matching found</returns>
        public RuleBase GetRule(string name)
        {
            foreach (RuleBase rb in _rules)
            {
                if (rb.Name == name)
                {
                    return rb;
                }
            }
            return null;
        }
        /// <summary>
        /// Get a token or charset by his name
        /// </summary>
        /// <param name="name">Token text or character set definition</param>
        /// <returns>The item or null if no matching found</returns>
        public IRuleItem GetItem(string name)
        {
            return _items.GetByName(name);
        }
        /// <summary>
        /// Links the item with the compilation event system
        /// </summary>
        /// <param name="item">Item to be linked</param>
        public void LinkItem(IRuleItem item)
        {
            // Only link once
            if (!item.Completed)
            {
                RuleBase rb = item as RuleBase;
                // This is only for rules in the main list
                if ((rb != null) && (_rules.Contains(rb)))
                {
                    rb.OnRootChanged += new RootChangedHandler(RootChanged);
                }
                if (item is ICompilableObjectGenerator)
                {
                    // When the item creates the compilable object
                    ((ICompilableObjectGenerator)item).OnCreateObject += new ObjectCreatedHandler(ObjectCreated);
                    // If the parsing fails and the object can be discarded
                    ((ICompilableObjectGenerator)item).OnDiscardObject += new ObjectCreatedHandler(DiscardObject);
                    // When the parsing is complete and the object is validated
                    ((ICompilableObjectGenerator)item).OnCompleteObject += new ObjectCreatedHandler(ObjectCompleted);
                }
                if (item is ItemBase)
                {
                    (item as ItemBase).OnIdChanged += new IdChangedEventHandler(ValidateItemBase);
                }
                item.Completed = true;
                if (item is RuleItemList)
                {
                    // Link also the content of the lists of items
                    foreach (IRuleItem ri in (item as RuleItemList))
                    {
                        LinkItem(ri);
                    }
                }
            }
        }
        /// <summary>
        /// Restores a copy of a IRuleItem using the canonical objects
        /// </summary>
        /// <param name="item">Item to restore</param>
        /// <returns>The canonical form of the item</returns>
        public IRuleItem RestoreItem(IRuleItem item)
        {
            if (item is RuleBase)
            {
                RuleBase rb = GetRule((item as RuleBase).Name);
                if (rb != null)
                {
                    return rb;
                }
                rb = item as RuleBase;
                for (int ix = 0; ix < rb.Count; ix++)
                {
                    rb[ix] = RestoreItem(rb[ix]);
                }
                return rb;
            }
            else if (item is RuleItemList)
            {
                RuleItemList ril = item as RuleItemList;
                for (int ix = 0; ix < ril.Count; ix++)
                {
                    ril[ix] = RestoreItem(ril[ix]);
                }
                return ril;
            }
            else if (item is ItemBase)
            {
                return GetItem(((ItemBase)item).ID);
            }
            else if (item is TokenList)
            {
                TokenList tl = item as TokenList;
                for (int ix = 0; ix < tl.Count; ix++)
                {
                    tl[ix] = GetItem(tl[ix].ID) as Token;
                }
                return tl;
            }
            else
            {
                return item;
            }
        }
        /// <summary>
        /// Performs the compilation of the objects
        /// </summary>
        /// <param name="rdr">TextReader with the source code to compile</param>
        /// <returns>The object generated by the current factory</returns>
        public ICompilableObject Build(TextReader rdr)
        {
            if (_factory == null)
            {
                throw new ObjectFactoryNotDefinedException();
            }
            if (_root == null)
            {
                throw new RootRuleNotDefinedException();
            }
            if (string.IsNullOrEmpty(_root.CompilableType))
            {
                throw new RootObjectNotDefinedException();
            }
            _current = null;
            _stack = new Stack<CompilerObject>();
            _factory.Init();
            Parser par = new Parser(rdr) { CaseSensitive = _caseSensitive };
            _root.Build(par);
            par.SkipWhitespaces();
            if (!par.EOF)
            {
                par.ThrowSyntaxException();
            }
            return _factory.Object.Simplify();
        }
        #endregion
        #region Event handlers
        /// <summary>
        /// Raised when one rule changes his Root property
        /// </summary>
        /// <param name="newroot">The rule which fired the event</param>
        private void RootChanged(RuleBase newroot)
        {
            if (newroot.Root && (newroot != _root))
            {
                if (_root != null)
                {
                    RuleBase rb = _root;
                    _root = null;
                    rb.Root = false;
                }
                _root = newroot;
            }
            else if ((!newroot.Root) && (newroot == _root))
            {
                _root = null;
            }
        }
        /// <summary>
        /// Fired when an item has created his ICompilableObject 
        /// </summary>
        /// <param name="owner">Object which has fired the event</param>
        private void ObjectCreated(ICompilableObjectGenerator owner)
        {
            if (_current != null)
            {
                // Push the current object in the stack
                _stack.Push(_current);
            }
            // Make the owner object the current one
            _current = new CompilerObject(owner, owner.Object);
            if (OnCompilationFeedback != null)
            {
                OnCompilationFeedback("ObjectCreated: " + _current.Object.ToString());
            }
        }
        /// <summary>
        /// Fired when the item has been correctly parsed and the object can be finished
        /// </summary>
        /// <param name="owner">Object which has fired the event</param>
        private void ObjectCompleted(ICompilableObjectGenerator owner)
        {
            if (owner is IRuleItem)
            {
                // Set the content of the object, which is the source text corresponding to the current item
                _current.Object.Text = ((IRuleItem)owner).GetText();
            }
            if (_stack.Count > 0)
            {
                try
                {
                    // Get the parent object from the current object
                    _current = _stack.Pop();
                    // Simplify if needed
                    ICompilableObject so = owner.Object.Simplify();
                    // Add the compiled object to the parent container
                    if (!_current.Object.AddItem(so))
                    {
                        throw new RejectedObjectException(_current.Object.Text, so.Text);
                    }
                    // Pass te current object to the item
                    _current.Owner.UpdateObject(_current.Object);
                    if (OnCompilationFeedback != null)
                    {
                        OnCompilationFeedback("ObjectCompleted: " + so.ToString());
                        OnCompilationFeedback("AddedTo: " + _current.Object.ToString());
                    }
                }
                catch (Exception ex)
                {
                    if (OnCompilationFeedback != null)
                    {
                        OnCompilationFeedback("Exception: " + ex.Message);
                    }
                    throw;
                }
            }
        }
        /// <summary>
        /// Fired when the parsing fails and the current object must be discarded
        /// </summary>
        /// <param name="owner">Object which has fired the event</param>
        private void DiscardObject(ICompilableObjectGenerator owner)
        {
            if (OnCompilationFeedback != null)
            {
                OnCompilationFeedback("DiscardObject: " + _current.Object.ToString());
            }
            if ((_current != null) && (_current.Owner == owner))
            {
                if (_stack.Count > 0)
                {
                    _current = _stack.Pop();
                }
                else
                {
                    _current = null;
                }
            }
        }
        #endregion
        #region Private Methods
        private void ValidateItemBase(string id)
        {
            if (GetItem(id) != null)
            {
                throw new Exception(Resources.ERR_VALUEINUSE);
            }
        }
        private void AddRule(RuleBase rule, List<RuleBase> rules)
        {
            if (rules.Contains(rule))
            {
                if (rules[rules.IndexOf(rule)] != rule)
                {
                    throw new DuplicateRuleNameException(rule.Name);
                }
                return;
            }
            rules.Add(rule);
            LinkItem(rule);            
        }
        private void RefreshRule(RuleBase newrule, RuleBase oldrule, List<RuleBase> rules)
        {
            if (rules.Contains(newrule))
            {
                if (rules[rules.IndexOf(newrule)] != newrule)
                {
                    rules[rules.IndexOf(newrule)] = newrule;
                    newrule.Completed = false;
                    LinkItem(newrule);
                    foreach (RuleBase rb in _rules)
                    {
                        if (rb != newrule)
                        {
                            rb.ChangeRule(oldrule, newrule, "");
                        }
                    }
                }
                return;
            }
        }
        #endregion
        #region Methods to parse BNF text files
        private string Identifier(Parser par)
        {
            string id = par.GetNext("^[a-zA-Z]");
            while (par.TryRegex("^[a-zA-Z0-9]"))
            {
                id += par.GetNext("^[a-zA-Z0-9]");
            }
            return id;
        }
        private string RName(Parser par)
        {
            if (par.TryParse("<<"))
            {
                if (string.IsNullOrEmpty(_rootname))
                {
                    par.Parse("<<");
                    _rootname = Identifier(par);
                    par.Parse(">>");
                    return _rootname;
                }
                else
                {
                    throw new DuplicatedRootRuleException();
                }
            }
            else
            {
                return RId(par);
            }
        }
        private string RId(Parser par)
        {
            par.Parse("<");
            string id = Identifier(par);
            par.Parse(">");
            return id;
        }
        private void RRules(Parser par)
        {
            Rule(par);
            while (!par.EOF)
            {
                Rule(par);
            }
        }
        private void Rule(Parser par)
        {
            string id = RName(par);
            par.Parse("::=");
            RuleBase rb = DefRule(par, id);
            par.Parse(";");
            AddRule(rb);
        }
        private RuleBase DefRule(Parser par, string name)
        {
            RuleBase rb = DefRuleCh(par, name);
            if (par.TryParse("|"))
            {
                par.Parse("|");
                AlternativeRules ar = new AlternativeRules(name);
                rb.Name = name + (ar.Count + 1).ToString();
                ar.Add(rb);
                rb = ar;
                rb.Add(DefRuleCh(par, name + (ar.Count + 1).ToString()));
                while(!par.EOF && par.TryParse("|"))
                {
                    par.Parse("|");
                    rb.Add(DefRuleCh(par, name + (ar.Count + 1).ToString()));
                }
            }
            return rb;
        }
        private RuleBase DefRuleCh(Parser par, string name)
        {
            IRuleItem item = Item(par);
            RuleChain rule = new RuleChain(name);
            rule.Add(item);
            return rule;
        }
        private IRuleItem Item(Parser par)
        {
            IRuleItem item = null;
            if (par.TryParse("["))
            {
                par.Parse("[");
                item = Item(par);
                if ((item is RuleBase) || (item is ItemBase))
                {
                    RuleItemList ril = new RuleItemList();
                    ril.Add(item);
                    item = ril;
                }
                item.Optional = true;
                par.Parse("]");
            }
            else if (par.TryParse("'"))
            {
                item = Token(par);
            }
            else if (par.TryParse("{"))
            {
                item = CharSet(par);
            }
            else
            {
                string id = RId(par);
                if (GetRule(id) != null)
                {
                    item = GetRule(id);
                }
                else
                {
                    item = new TempRule(id);
                }
            }
            if (par.TryParse("'") || par.TryParse("{") || par.TryParse("<") || par.TryParse("["))
            {
                RuleItemList ril = new RuleItemList();
                LinkItem(item);
                ril.Add(item);
                item = ril;
                while (par.TryParse("'") || par.TryParse("{") || par.TryParse("<") || par.TryParse("["))
                {
                    IRuleItem nxti = Item(par);
                    if ((nxti is RuleItemList) && (nxti.Optional == item.Optional))
                    {
                        foreach (IRuleItem lc in nxti.Childs)
                        {
                            ril.Add(lc);
                        }
                    }
                    else
                    {
                        ril.Add(nxti);
                    }
                }
            }
            LinkItem(item);
            return item;
        }
        private string TokenId(Parser par)
        {
            string token = "";
            par.Parse("'");
            if (par.TryParse("\\'"))
            {
                par.Parse("\\'");
                token += "'";
            }
            else if (!par.TryParse("'"))
            {
                token += par.GetNext(".");
            }
            token += RToken(par);
            par.Parse("'");
            return token;
        }
        private string RToken(Parser par)
        {
            string rtoken = "";
            while (!par.TryParse("'"))
            {
                if (par.TryParse("\\'"))
                {
                    par.Parse("\\'");
                    rtoken += "'";
                }
                else if (!par.TryParse("'"))
                {
                    rtoken += par.GetNext(".");
                }
            }
            return rtoken;
        }
        private IRuleItem Token(Parser par)
        {
            string token = TokenId(par);
            Token t = _items.AddNewToken(new Token(token));            
            if (par.TryParse(","))
            {
                TokenList tl = new TokenList();
                LinkItem(t);
                tl.Add(t);
                while (par.TryParse(","))
                {
                    par.Parse(",");
                    token = TokenId(par);
                    t = _items.AddNewToken(new Token(token));
                    LinkItem(t);
                    tl.Add(t);
                }
                return tl;
            }
            else
            {
                return t;
            }
        }
        private IRuleItem CharSet(Parser par)
        {
            string charset = "";
            par.Parse("{");
            if (par.TryParse("\\}"))
            {
                par.Parse("\\}");
                charset += "}";
            }
            else if (!par.TryParse("}"))
            {
                charset += par.GetNext(".");
            }
            charset += CharList(par);
            par.Parse("}");
            return _items.AddNewCharacterSet(new CharacterSet(charset));
        }
        private string CharList(Parser par)
        {
            string charset = "";
            while (!par.TryParse("}"))
            {
                if (par.TryParse("\\}"))
                {
                    par.Parse("\\}");
                    charset += "}";
                }
                else if (!par.TryParse("}"))
                {
                    charset += par.GetNext(".");
                }
            }
            return charset;
        }
        #endregion
    }
}
