﻿using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Drawing;
using System.ComponentModel;
using BNFUP.Interfaces;
using BNFUP.Attributes;
using BNFUP.Properties;
using BNFUP.Exceptions;
using BNFUP.Controls;

namespace BNFUP.Rules
{
    [Serializable]
    public class TokenList : List<Token>, IRuleItem, ICompilableObjectGenerator, IItemDrawing, IEditableItem
    {
        private int _color = Color.Black.ToArgb();
        private string _text;
        private bool _completed = false;
        [NonSerialized]
        protected ICompilableObject _object = null;
        [NonSerialized]
        protected ICompilableObjectFactory _factory = null;
        public TokenList()
        {
        }
        [Browsable(false)]
        public new int Capacity
        {
            get
            {
                return base.Capacity;
            }
            set
            {
                base.Capacity = value;
            }
        }
        [Browsable(false)]
        public new int Count
        {
            get
            {
                return base.Count;
            }
        }
        private string Options
        {
            get
            {
                List<string> opts = new List<string>();
                foreach (Token t in this)
                {
                    if (t.Name == ",")
                    {
                        opts.Add(",");
                    }
                    else
                    {
                        opts.Add(t.Name);
                    }
                }
                return string.Join(",", opts.ToArray());
            }
        }
        #region IRuleItem
        [Browsable(false)]
        public bool Completed
        {
            get
            {
                return _completed;
            }
            set
            {
                _completed = value;
                if (!_completed)
                {
                    OnCompleteObject = null;
                    OnCreateObject = null;
                    OnDiscardObject = null;
                }
            }
        }
        [CategoryGlobal("CAT_SYNTAX")]
        [DisplayNameGlobal("DIS_OPTIONAL")]
        [DescriptionGlobal("DES_OPTIONAL")]
        public bool Optional { get; set; }
        [Browsable(false)]
        public string Description
        {
            get
            {
                string s = "";
                string comma = "";
                foreach (Token t in this)
                {
                    s += comma + t.ToString();
                    comma = ", ";
                }
                return Resources.DES_TOKENLIST + ": " + s;
            }
        }
        [Browsable(false)]
        public IEnumerable<IRuleItem> Childs
        {
            get
            {
                foreach (Token t in this)
                {
                    yield return t;
                }
            }
        }
        public string GetText()
        {
            return _text;
        }
        public void UpdateRule(string name, RuleBase rule, string current) { }
        public void ChangeRule(RuleBase oldrule, RuleBase newrule, string current) { }
        public void DeleteRule(RuleBase rule, string current) { }
        public bool Matches(Parser parser)
        {
            foreach (Token t in this)
            {
                if (t.Matches(parser))
                {
                    return true;
                }
            }
            return false;
        }
        public void Build(Parser parser)
        {
            _text = "";
            foreach (Token t in this)
            {
                if (t.Matches(parser))
                {
                    t.Build(parser);
                    _text = t.GetText();
                    break;
                }
            }
            if (string.IsNullOrEmpty(_text))
            {
                parser.ThrowSyntaxException();
            }
            if (_factory != null)
            {
                if (!string.IsNullOrEmpty(CompilableType))
                {
                    _object = _factory.CreateObject(CompilableType, this);
                    if (_object == null)
                    {
                        throw new UnknownObjectKeyException(CompilableType);
                    }
                    OnCreateObject?.Invoke(this);
                    OnCompleteObject?.Invoke(this);
                }
            }
            else
            {
                throw new ObjectFactoryNotDefinedException();
            }
        }
        #endregion
        #region ICompilableObjectGenerator
        [CategoryGlobal("CAT_COMPILER")]
        [DisplayNameGlobal("DIS_OBJTYPE")]
        [DescriptionGlobal("DES_OBJTYPE")]
        public string CompilableType { get; set; }
        [Browsable(false)]
        public ICompilableObjectFactory Factory
        {
            get
            {
                return _factory;
            }
            set
            {
                _factory = value;
                foreach (Token t in this)
                {
                    t.Factory = value;
                }
            }
        }
        [Browsable(false)]
        public ICompilableObject Object
        {
            get
            {
                return _object;
            }
        }
        public void UpdateObject(ICompilableObject obj)
        {
            _object = obj;
        }
        [Browsable(false)]
        public event ObjectCreatedHandler OnCreateObject = null;
        [Browsable(false)]
        public event ObjectCreatedHandler OnDiscardObject = null;
        [Browsable(false)]
        public event ObjectCreatedHandler OnCompleteObject = null;
        #endregion
        #region IItemDrawing
        [CategoryGlobal("CAT_DISPLAY")]
        [DisplayNameGlobal("DIS_COLOR")]
        [DescriptionGlobal("DES_COLOR")]
        public Color ForeColor
        {
            get
            {
                return Color.FromArgb(_color);
            }
            set
            {
                _color = value.ToArgb();
            }
        }
        public SizeF Measure(Graphics gr, Font fextra, Font fmain)
        {
            SizeF szn = gr.MeasureString(Options, fmain);
            float wo = 0f;
            if (Optional)
            {
                wo = gr.MeasureString("[ ", fextra).Width + gr.MeasureString(" ]", fextra).Width;
            }
            return new SizeF(szn.Width + wo, szn.Height);
        }
        public void Draw(Graphics gr, float x, float y, Font fextra, Font fmain)
        {
            Brush bc = new SolidBrush(ForeColor);
            if (Optional)
            {
                gr.DrawString("[ ", fextra, Brushes.Black, x, y);
                x += gr.MeasureString("[ ", fextra).Width;
            }
            gr.DrawString(Options, fmain, bc, x, y);
            x += gr.MeasureString(Options, fmain).Width;
            if (Optional)
            {
                gr.DrawString(" ]", fextra, Brushes.Black, x, y);
            }
            bc.Dispose();
        }
        #endregion
        #region IEditableItem
        [Browsable(false)]
        public bool CanCopy
        {
            get
            {
                foreach (Token t in this)
                {
                    if (!t.CanCopy)
                    {
                        return false;
                    }
                }
                return true;
            }
        }
        [Browsable(false)]
        public int ItemIndex { get; set; }
        public IEnumerable<ToolStripItem> EditOptions(IEditableItem parent, IEditableItem gparent, IRuleItemClipboard clipboard)
        {
            ToolStripMenuItem ti = new ToolStripMenuItem(Resources.OPT_NEW);
            ti.Tag = EditActions.New;
            yield return ti;
            ti = new ToolStripMenuItem(Resources.OPT_DELETE);
            ti.Tag = EditActions.Delete;
            yield return ti;
            ti = new ToolStripMenuItem(Resources.OPT_FREE);
            ti.Tag = EditActions.Free;
            yield return ti;
            bool cc = CanCopy;
            if (cc && (clipboard != null))
            {
                yield return new ToolStripSeparator();
                ti = new ToolStripMenuItem(Resources.OPT_CUT);
                ti.Tag = EditActions.Cut;
                yield return ti;
                ti = new ToolStripMenuItem(Resources.OPT_COPY);
                ti.Tag = EditActions.Copy;
                yield return ti;
            }
            if ((clipboard != null) && clipboard.ContainsData)
            {
                IRuleItem ir = clipboard.GetData();
                if ((ir is Token) || (ir is TokenList))
                {
                    if (!cc)
                    {
                        yield return new ToolStripSeparator();
                    }
                    ti = new ToolStripMenuItem(Resources.OPT_PASTE);
                    ti.Tag = EditActions.Paste;
                    yield return ti;
                }
            }
            yield return new ToolStripSeparator();
            ti = new ToolStripMenuItem(Resources.OPT_ENLIST);
            ti.Tag = EditActions.Enlist;
            yield return ti;
            if (gparent != null)
            {
                ti = new ToolStripMenuItem(Resources.OPT_EXTRACT);
                ti.Tag = EditActions.Extract;
                yield return ti;
            }
            yield return new ToolStripSeparator();
            ti = new ToolStripMenuItem(Resources.OPT_MOVEUP);
            ti.Tag = EditActions.Up;
            yield return ti;
            ti = new ToolStripMenuItem(Resources.OPT_MOVEDOWN);
            ti.Tag = EditActions.Down;
            yield return ti;
            yield break;
        }
        public bool PerformAction(ToolStripItem action, IEditableItem parent, IEditableItem gparent, RuleTable rules, IRuleItemClipboard clipboard)
        {
            if (action.Tag is EditActions)
            {
                EditActions ca = (EditActions)action.Tag;
                switch (ca)
                {
                    case EditActions.Delete:
                        parent.RemoveChild(ItemIndex);
                        break;
                    case EditActions.Cut:
                        parent.RemoveChild(ItemIndex);
                        clipboard.SetData(this);
                        break;
                    case EditActions.Copy:
                        clipboard.SetData(this);
                        break;
                    case EditActions.Paste:
                        if (clipboard.ContainsData)
                        {
                            IRuleItem ir = clipboard.GetData();
                            if (ir is ICompilableObjectGenerator)
                            {
                                ((ICompilableObjectGenerator)ir).Factory = _factory;
                            }
                            if (ir is Token)
                            {
                                Token t = ir as Token;
                                t = rules.GetItem(t.ID) as Token;
                                Add(t);
                            }
                            else if (ir is TokenList)
                            {
                                foreach (Token t in (ir as TokenList))
                                {
                                    Add(rules.GetItem(t.ID) as Token);
                                }
                            }
                        }
                        break;
                    case EditActions.Enlist:
                        RuleItemList l = new RuleItemList();
                        l.Factory = _factory;
                        l.Add(this);
                        parent.ReplaceChild(ItemIndex, l);
                        break;
                    case EditActions.Extract:
                        if (gparent != null)
                        {
                            gparent.InsertAfter(parent.ItemIndex, this);
                            parent.RemoveChild(ItemIndex);
                        }
                        break;
                    case EditActions.Free:
                        int ix = ItemIndex + 1;
                        foreach (Token t in this)
                        {
                            parent.InsertAfter(ix++, t);
                        }
                        parent.RemoveChild(ItemIndex);
                        break;
                    case EditActions.Up:
                        parent.MoveUp(ItemIndex);
                        break;
                    case EditActions.Down:
                        parent.MoveDown(ItemIndex);
                        break;
                    case EditActions.New:
                        dlgItems dlg = new dlgItems();
                        dlg.ListOwner = this;
                        dlg.Rules = rules;
                        if (dlg.ShowDialog() == DialogResult.OK)
                        {
                            Token t = dlg.SelectedObject as Token;
                            t.Factory = _factory;
                            if (t != null)
                            {
                                AddChild(t);
                                rules.AddItem(t);
                            }
                        }
                        break;
                }
            }
            return false;
        }
        public void AddChild(IRuleItem item)
        {
            Token t = item as Token;
            if (t != null)
            {
                Add(t);
            }
        }
        public void RemoveChild(int item)
        {
            RemoveAt(item);
        }
        public void ReplaceChild(int item1, IRuleItem item2)
        {
            if (item2 is Token)
            {
                this[item1] = item2 as Token;
            }
        }
        public void InsertAfter(int item1, IRuleItem item2)
        {
            if (item2 is Token)
            {
                Insert(item1, item2 as Token);
            }
        }
        public void MoveUp(int item)
        {         
            if (item > 0)
            {
                Token tmp = this[item - 1];
                this[item - 1] = this[item];
                this[item] = tmp;
            }
        }
        public void MoveDown(int item)
        {
            if (item < Count - 1)
            {
                Token tmp = this[item + 1];
                this[item + 1] = this[item];
                this[item] = tmp;
            }
        }
        public bool ValidateType(Type t, bool generic)
        {
            return t == typeof(Token);
        }
        #endregion
    }
}
