﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Drawing;
using System.Drawing.Imaging;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using BNFUP.Rules;
using CommonObjects.Interfaces;
using CellularAutomata.Interfaces;
using CellularAutomata.Properties;
using CellularAutomata.Controls;
using CellularAutomata.Expressions;

namespace CellularAutomata.Automata
{
    [Serializable]
    public class Grid2DCA : ICANetwork
    {
        [NonSerialized]
        private ICAOptimizer _optimizer = null;
        [NonSerialized]
        private Grid2DCAEditor _editor = null;
        [NonSerialized]
        private ToolStripComboBox _cbCells = null;
        [NonSerialized]
        private ToolStripTextBox _textWidth = null;
        [NonSerialized]
        private ToolStripTextBox _textHeight = null;
        [NonSerialized]
        private ICACellProvider _cellProvider = null;
        private ICACell[] _cells = null;
        private int[] _initialState = null;
        private int[] _neighborhood = null;
        private bool _periodic = true;
        private int _width = 100;
        private int _height = 100;
        private int _currentCell = 0;
        private int _iteratorIx = -1;
        private int _step = 0;
        private int _zoom = 1;
        private Stack<int> _iterators = new Stack<int>();
        private List<int> _colors = new List<int>();

        public Grid2DCA()
        {
        }
        public Size Size
        {
            get
            {
                return new Size(_width, _height);
            }
            set
            {
                if ((value.Width >= 20) && (value.Width <= 1000) && (value.Height >= 20) && (value.Height <= 1000))
                {
                    _width = value.Width;
                    _height = value.Height;
                    CreateNetwork();
                    if (_editor != null)
                    {
                        _editor.Length = value;
                    }
                }
            }
        }
        public override string ToString()
        {
            return UIResources.CAN_GRID2D;
        }
        #region IValueProvider
        public IVariantValue GetVarValue(string name)
        {
            return VariableProvider.GetProperty(name).Value;
        }
        public IVariantValue GetPropertyValue(string name)
        {
            ICAState st = StateProvider.GetState(_cells[_currentCell].ActualState);
            return st.GetVarValue(name);
        }
        public IVariantValue GetIteratorPropertyValue(string name)
        {
            return GetIndexedPropertyValue(_iteratorIx, name);
        }
        public IVariantValue GetIndexedPropertyValue(int ix, string name)
        {
            if ((ix < 0) || (ix >= _cells[_currentCell].NeighborCount))
            {
                throw new Exception(string.Format(UIResources.ERR_BADINDEX, ix, name));
            }
            ICAState st = StateProvider.GetState(_cells[_cells[_currentCell].GetNeighbour(ix)].ActualState);
            return st.GetVarValue(name);
        }
        public void BeginIterator()
        {
            if (_iteratorIx >= 0)
            {
                _iterators.Push(_iteratorIx);
            }
            _iteratorIx = 0;
        }
        public bool Next()
        {
            _iteratorIx++;
            return _iteratorIx < _cells[_currentCell].NeighborCount;
        }
        public void EndIterator()
        {
            if (_iterators.Count > 0)
            {
                _iteratorIx = _iterators.Pop();
            }
            else
            {
                _iteratorIx = -1;
            }
        }
        #endregion
        #region ICANetwork
        public ICAOptimizer Optimizer
        {
            get
            {
                return _optimizer;
            }
            set
            {
                _optimizer = value;
            }
        }
        public ICAStateProvider StateProvider { get; set; }
        public IPropertyProvider VariableProvider { get; set; }
        public ICACellProvider CellProvider
        {
            get
            {
                return _cellProvider;
            }
            set
            {
                _cellProvider = value;
                if (_cbCells != null)
                {
                    _cbCells.Items.Clear();
                    if (value != null)
                    {
                        foreach (Type t in value.CellTypes)
                        {
                            _cbCells.Items.Add(t.GetConstructor(Type.EmptyTypes).Invoke(null));
                        }
                        if (_cbCells.Items.Count > 0)
                        {
                            _cbCells.SelectedIndex = 0;
                        }
                    }
                }
            }
        }
        public Control Editor
        {
            get
            {
                if (_editor == null)
                {
                    _editor = new Grid2DCAEditor()
                    {
                        Length = new Size(_width, _height),
                        Neighborhood = _neighborhood
                    };
                    _editor.OnChangeNeighborhood += new EventHandler(EditorChange);
                }
                return _editor;
            }
        }
        public int ActualStep
        {
            get
            {
                return _step;
            }
        }
        public Bitmap ActualImage
        {
            get
            {
                return GetImage();
            }
        }
        public void ConfigureToolBar(ToolStrip tb, int ix, bool remove)
        {
            if (remove)
            {
                for (int i = 0; i < 10; i++)
                {
                    tb.Items.RemoveAt(ix);
                }
            }
            else
            {
                ToolStripLabel lb = new ToolStripLabel(UIResources.LAB_WIDTH);
                tb.Items.Insert(ix++, lb);
                _textWidth = new ToolStripTextBox();
                _textWidth.Width = 40;
                _textWidth.Text = _width.ToString();
                tb.Items.Insert(ix++, _textWidth);
                ToolStripButton btn = new ToolStripButton(Resources.BTN_CHANGE);
                btn.ImageTransparentColor = Color.Magenta;
                btn.ToolTipText = UIResources.TTP_CHANGEWIDTH;
                btn.Click += new EventHandler(WidthChanged);
                tb.Items.Insert(ix++, btn);
                lb = new ToolStripLabel(UIResources.LAB_HEIGHT);
                tb.Items.Insert(ix++, lb);
                _textHeight = new ToolStripTextBox();
                _textHeight.Width = 40;
                _textHeight.Text = _height.ToString();
                tb.Items.Insert(ix++, _textHeight);
                btn = new ToolStripButton(Resources.BTN_CHANGE);
                btn.ImageTransparentColor = Color.Magenta;
                btn.ToolTipText = UIResources.TTP_CHANGEHEIGHT;
                btn.Click += new EventHandler(HeightChanged);
                tb.Items.Insert(ix++, btn);
                btn = new ToolStripButton(Resources.BTN_INITIAL);
                btn.ImageTransparentColor = Color.Magenta;
                btn.ToolTipText = UIResources.TTP_INITIALIZE;
                btn.Click += new EventHandler(Initialize);
                tb.Items.Insert(ix++, btn);
                lb = new ToolStripLabel(UIResources.LAB_CELLTYPE);
                tb.Items.Insert(ix++, lb);
                _cbCells = new ToolStripComboBox();
                _cbCells.Width = 150;
                _cbCells.Sorted = true;
                _cbCells.DropDownStyle = ComboBoxStyle.DropDownList;
                CellProvider = _cellProvider;
                if (_cbCells.Items.Count > 0)
                {
                    _cbCells.SelectedIndex = 0;
                }
                tb.Items.Insert(ix++, _cbCells);
                btn = new ToolStripButton(Resources.BTN_PERIODIC);
                btn.ImageTransparentColor = Color.Magenta;
                btn.ToolTipText = UIResources.TTP_PERIODIC;
                btn.CheckOnClick = true;
                btn.Checked = true;
                btn.Click += new EventHandler(SetPeriodic);
                tb.Items.Insert(ix++, btn);
            }
        }
        public void Initialize()
        {
            CreateNetwork();
            InitializeNeighborhood();
            Reset();
        }
        public Bitmap Reset()
        {
            _iterators.Clear();
            _iteratorIx = -1;
            _step = 0;
            _currentCell = 0;
            InitializeState();
            _colors.Clear();
            for (int ix = 0; ix < StateProvider.Count; ix++)
            {
                _colors.Add(StateProvider.GetState(ix).StateColor.ToArgb());
            }
            MemoryStream ms = new MemoryStream(Resources.expressions);
            RuleTable rt = RuleTable.FromStream(ms);
            rt.Factory = new ExpressionFactory()
            {
                ValueProvider = this
            };
            for (int ix = 0; ix < StateProvider.Count; ix++)
            {
                StateProvider.GetState(ix).CompileTransitions(rt, StateProvider);
            }
            return GetImage();
        }
        public Bitmap Step()
        {
            Bitmap bmp = null;
            if (_step > 0)
            {
                for (_currentCell = 0; _currentCell < _width * _height; _currentCell++)
                {
                    int stix = _cells[_currentCell].ActualState;
                    ICAState st = StateProvider.GetState(stix);
                    if (_optimizer != null)
                    {
                        int[] nh = _cells[_currentCell].GetNeighborhood();
                        for (int ix = 0; ix < nh.Length; ix++)
                        {
                            nh[ix] = _cells[nh[ix]].ActualState;
                        }
                        int nwst = _optimizer.GetTransition(stix, nh);
                        if (nwst >= 0)
                        {
                            _cells[_currentCell].NextState = nwst;
                        }
                        else
                        {
                            nwst = st.CheckTransitions();
                            if (nwst >= 0)
                            {
                                _cells[_currentCell].NextState = nwst;
                                _optimizer.AddKnownTransition(stix, nh, nwst);
                            }
                            else
                            {
                                _cells[_currentCell].NextState = _cells[_currentCell].ActualState;
                            }
                        }
                    }
                    else
                    {
                        int nwst = st.CheckTransitions();
                        if (nwst >= 0)
                        {
                            _cells[_currentCell].NextState = nwst;
                        }
                        else
                        {
                            _cells[_currentCell].NextState = _cells[_currentCell].ActualState;
                        }
                    }
                }
            }
            bmp = GetImage();
            _step++;
            return bmp;
        }
        public Bitmap ZoomIn()
        {
            if (_zoom < 32)
            {
                _zoom <<= 1;
            }
            return GetImage();
        }
        public Bitmap ZoomOut()
        {
            if (_zoom > 1)
            {
                _zoom >>= 1;
            }
            return GetImage();
        }
        #endregion
        private Bitmap GetImage()
        {
            int w = _width * _zoom;
            int h = _height * _zoom;
            Bitmap bmp = new Bitmap(w, h);
            BitmapData bdata = null;
            try
            {

                bdata = bmp.LockBits(new Rectangle(0, 0, w, h), ImageLockMode.ReadWrite, PixelFormat.Format32bppRgb);
                int[] data = new int[w * h];
                int pd = 0;
                int iyw = 0;
                for (int iy = 0; iy < _height; iy++)
                {                    
                    int pd0 = pd;
                    for (int ix = 0; ix < _width; ix++)
                    {
                        for (int ih = 0; ih < _zoom; ih++)
                        {
                            data[pd++] = _colors[_cells[ix + iyw].GetState(_step)];
                        }
                    }
                    for (int iv = 1; iv < _zoom; iv++)
                    {
                        int pdz = pd0;
                        for (int ix = 0; ix < w; ix++)
                        {
                            data[pd++] = data[pdz++];
                        }
                    }
                    iyw += _width;
                }
                Marshal.Copy(data, 0, bdata.Scan0, w * h);
                return bmp;
            }
            finally
            {
                if (bdata != null)
                {
                    bmp.UnlockBits(bdata);
                }
            }
        }
        private void CreateNetwork()
        {
            _cells = new ICACell[_width * _height];
            if ((_initialState == null) || (_initialState.Length != _width * _height))
            {
                _initialState = new int[_width * _height];
            }
            for (int ix = 0; ix < _width * _height; ix++)
            {
                _cells[ix] = _cellProvider.GetCell(_cbCells.SelectedItem.GetType(), 0);
            }
        }
        private void InitializeState()
        {
            for (int ix = 0; ix < _width * _height; ix++)
            {
                _cells[ix].ActualState = _initialState[ix];
            }
        }
        private void InitializeNeighborhood()
        {
            if ((_cells != null) && (_neighborhood != null))
            {
                int cntr = (_width - 1) / 2;
                int cntl = (_width - 1) - cntr;
                for (int iy = 0; iy < _height; iy++)
                {
                    for (int ix = 0; ix < _width; ix++)
                    {
                        List<int> nc = new List<int>();
                        for (int n = 0; n < _neighborhood.Length; n++)
                        {
                            if (_neighborhood[n] != 0)
                            {
                                int x = 0;
                                int y = 0;
                                if (_neighborhood[n] < 0)
                                {
                                    y = (_neighborhood[n] - cntl) / _width;                                    
                                }
                                else
                                {
                                    y = (_neighborhood[n] + cntr) / _width;
                                }
                                x = _neighborhood[n] - (y * _width);
                                if (_periodic)
                                {
                                    if (ix + x < 0)
                                    {
                                        if (iy + y < 0)
                                        {
                                            nc.Add((_width + (ix + x)) + _width * (_height + (iy + y)));
                                        }
                                        else if (iy + y >= _height)
                                        {
                                            nc.Add((_width + (ix + x)) + _width * ((iy + y) - _height));
                                        }
                                        else
                                        {
                                            nc.Add((_width + (ix + x)) + _width * (iy + y));
                                        }
                                    }
                                    else if (ix + x >= _width)
                                    {
                                        if (iy + y < 0)
                                        {
                                            nc.Add(((ix + x) - _width) + _width * (_height + (iy + y)));
                                        }
                                        else if (iy + y >= _height)
                                        {
                                            nc.Add(((ix + x) - _width) + _width * ((iy + y) - _height));
                                        }
                                        else
                                        {
                                            nc.Add(((ix + x) - _width) + _width * (iy + y));
                                        }
                                    }
                                    else
                                    {
                                        if (iy + y < 0)
                                        {
                                            nc.Add((ix + x) + _width * (_height + (iy + y)));
                                        }
                                        else if (iy + y >= _height)
                                        {
                                            nc.Add((ix + x) + _width * ((iy + y) - _height));
                                        }
                                        else
                                        {
                                            nc.Add((ix + x) + _width * (iy + y));
                                        }
                                    }
                                }
                                else
                                {
                                    if ((ix + x >= 0) && (ix + x < _width) && (iy + y >= 0) && (iy + y < _height))
                                    {
                                        nc.Add((ix + x) + _width * (iy + y));
                                    }
                                }
                            }
                        }
                        _cells[ix + iy * _width].AddNeighbours(nc.ToArray());
                    }
                }
            }
        }
        private void EditorChange(object sender, EventArgs e)
        {
            _neighborhood = _editor.Neighborhood;
            InitializeNeighborhood();
        }
        private void WidthChanged(object sender, EventArgs e)
        {
            int w = 0;
            if (int.TryParse(_textWidth.Text, out w))
            {
                Size = new Size(w, _height);
            }
            _textWidth.Text = _width.ToString();
        }
        private void HeightChanged(object sender, EventArgs e)
        {
            int h = 0;
            if (int.TryParse(_textHeight.Text, out h))
            {
                Size = new Size(_width, h);
            }
            _textHeight.Text = _height.ToString();
        }
        private void SetPeriodic(object sender, EventArgs e)
        {
            ToolStripButton btn = sender as ToolStripButton;
            if (btn != null)
            {
                _periodic = btn.Checked;
                if (_periodic)
                {
                    btn.ToolTipText = UIResources.TTP_PERIODIC;
                }
                else
                {
                    btn.ToolTipText = UIResources.TTP_FINITE;
                }
            }
        }
        private void Initialize(object sender, EventArgs e)
        {
            GridEditor dlg = new GridEditor();
            dlg.StateProvider = StateProvider;
            dlg.ItemSize = new Size(_width, _height);
            if (_initialState != null)
            {
                int[] row = new int[_width];
                for (int iy = 0; iy < _height; iy++)
                {
                    for (int ix = 0; ix < _width; ix++)
                    {
                        row[ix] = _initialState[ix + iy * _width];
                    }
                    dlg.SetImageRow(iy, row);
                }
            }
            if (dlg.ShowDialog() == DialogResult.OK)
            {
                _initialState = new int[_width * _height];
                for (int iy = 0; iy < _height; iy++)
                {
                    int[] row = dlg.GetImageRow(iy);
                    for (int ix = 0; ix < _width; ix++)
                    {
                        _initialState[ix + iy * _width] = row[ix];
                    }
                }
                InitializeState();
            }
        }
    }
}
