﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Drawing;
using System.Drawing.Imaging;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using BNFUP.Rules;
using CommonObjects.Interfaces;
using CellularAutomata.Interfaces;
using CellularAutomata.Properties;
using CellularAutomata.Controls;
using CellularAutomata.Expressions;

namespace CellularAutomata.Automata
{
    [Serializable]
    public class Linear1DCA : ICANetwork
    {
        [NonSerialized]
        private ICAOptimizer _optimizer = null;
        [NonSerialized]
        private Linear1DCAEditor _editor = null;
        [NonSerialized]
        private ToolStripComboBox _cbCells = null;
        [NonSerialized]
        private ToolStripTextBox _textWidth = null;
        [NonSerialized]
        private ToolStripTextBox _textHistory = null;
        [NonSerialized]
        private int[,] _memory = null;
        private int _memoryPos = 0;
        private ICACellProvider _cellProvider = null;
        private ICACell[] _cells = null;
        private int[] _initialState = null;
        private int[] _neighborhood = null;
        private bool _periodic = true;
        private int _width = 100;
        private int _history = 100;
        private int _currentCell = 0;
        private int _iteratorIx = -1;
        private int _step = 0;
        private int _zoom = 1;
        private Stack<int> _iterators = new Stack<int>();
        private List<int> _colors = new List<int>();        

        public Linear1DCA()
        {            
        }
        public int Width
        {
            get
            {
                return _width;
            }
            set
            {
                if ((value >= 20) && (value <= 1000))
                {
                    _width = value;
                    CreateNetwork();
                    if (_editor != null)
                    {
                        _editor.Length = _width;
                    }
                }            
            }
        }
        public int History
        {
            get
            {
                return _history;
            }
            set
            {
                if ((value >= 2) && (value <= 1000))
                {
                    _history = value;
                }
            }
        }
        public override string ToString()
        {
            return UIResources.CAN_LINEAR1D;
        }
        #region IValueProvider
        public IVariantValue GetVarValue(string name)
        {
            return VariableProvider.GetProperty(name).Value;
        }
        public IVariantValue GetPropertyValue(string name)
        {
            ICAState st = StateProvider.GetState(_cells[_currentCell].ActualState);
            return st.GetVarValue(name);
        }
        public IVariantValue GetIteratorPropertyValue(string name)
        {
            return GetIndexedPropertyValue(_iteratorIx, name);
        }
        public IVariantValue GetIndexedPropertyValue(int ix, string name)
        {
            if ((ix < 0) || (ix >= _cells[_currentCell].NeighborCount))
            {
                throw new Exception(string.Format(UIResources.ERR_BADINDEX, ix, name));
            }
            ICAState st = StateProvider.GetState(_cells[_cells[_currentCell].GetNeighbour(ix)].ActualState);
            return st.GetVarValue(name);
        }
        public void BeginIterator()
        {
            if (_iteratorIx >= 0)
            {
                _iterators.Push(_iteratorIx);
            }
            _iteratorIx = 0;
        }
        public bool Next()
        {
            _iteratorIx++;
            return _iteratorIx < _cells[_currentCell].NeighborCount;
        }
        public void EndIterator()
        {
            if (_iterators.Count > 0)
            {
                _iteratorIx = _iterators.Pop();
            }
            else
            {
                _iteratorIx = -1;                
            }
        }
        #endregion
        #region ICANetwork
        public ICAOptimizer Optimizer
        {
            get
            {
                return _optimizer; 
            }
            set
            {
                _optimizer = value;
            }
        }
        public ICAStateProvider StateProvider { get; set; }
        public IPropertyProvider VariableProvider { get; set; }
        public ICACellProvider CellProvider
        {
            get
            {
                return _cellProvider;
            }
            set
            {
                _cellProvider = value;
                if (_cbCells != null)
                {
                    _cbCells.Items.Clear();
                    if (value != null)
                    {
                        foreach (Type t in value.CellTypes)
                        {
                            _cbCells.Items.Add(t.GetConstructor(Type.EmptyTypes).Invoke(null));
                        }
                        if (_cbCells.Items.Count > 0)
                        {
                            _cbCells.SelectedIndex = 0;
                        }
                    }
                }
            }
        }
        public Control Editor
        {
            get
            {
                if (_editor == null)
                {
                    _editor = new Linear1DCAEditor()
                    {
                        Length = _width,
                        Neighborhood = _neighborhood
                    };
                    _editor.OnChangeNeighborhood += new EventHandler(EditorChange);
                }
                return _editor;
            }
        }
        public int ActualStep
        {
            get
            {
                return _step;
            }
        }
        public Bitmap ActualImage
        {
            get
            {
                return GetImage();
            }
        }
        public void ConfigureToolBar(ToolStrip tb, int ix, bool remove)
        {
            if (remove)
            {
                for (int i = 0; i < 10; i++)
                {
                    tb.Items.RemoveAt(ix);
                }
            }
            else
            {
                ToolStripLabel lb = new ToolStripLabel(UIResources.LAB_WIDTH);                
                tb.Items.Insert(ix++, lb);
                _textWidth = new ToolStripTextBox();
                _textWidth.Width = 40;
                _textWidth.Text = _width.ToString();
                tb.Items.Insert(ix++, _textWidth);
                ToolStripButton btn = new ToolStripButton(Resources.BTN_CHANGE);
                btn.ImageTransparentColor = Color.Magenta;
                btn.ToolTipText = UIResources.TTP_CHANGEWIDTH;
                btn.Click += new EventHandler(WidthChanged);
                tb.Items.Insert(ix++, btn);
                lb = new ToolStripLabel(UIResources.LAB_HISTORY);
                tb.Items.Insert(ix++, lb);
                _textHistory = new ToolStripTextBox();
                _textHistory.Width = 40;
                _textHistory.Text = _history.ToString();
                tb.Items.Insert(ix++, _textHistory);
                btn = new ToolStripButton(Resources.BTN_CHANGE);
                btn.ImageTransparentColor = Color.Magenta;
                btn.ToolTipText = UIResources.TTP_CHANGEHIST;
                btn.Click += new EventHandler(HistoryChanged);
                tb.Items.Insert(ix++, btn);
                btn = new ToolStripButton(Resources.BTN_INITIAL);
                btn.ImageTransparentColor = Color.Magenta;
                btn.ToolTipText = UIResources.TTP_INITIALIZE;
                btn.Click += new EventHandler(Initialize);
                tb.Items.Insert(ix++, btn);
                lb = new ToolStripLabel(UIResources.LAB_CELLTYPE);
                tb.Items.Insert(ix++, lb);
                _cbCells = new ToolStripComboBox();
                _cbCells.Width = 150;
                _cbCells.Sorted = true;
                _cbCells.DropDownStyle = ComboBoxStyle.DropDownList;
                CellProvider = _cellProvider;
                if (_cbCells.Items.Count > 0)
                {
                    _cbCells.SelectedIndex = 0;
                }
                tb.Items.Insert(ix++, _cbCells);
                btn = new ToolStripButton(Resources.BTN_PERIODIC);
                btn.ImageTransparentColor = Color.Magenta;
                btn.ToolTipText = UIResources.TTP_PERIODIC;
                btn.CheckOnClick = true;
                btn.Checked = true;
                btn.Click += new EventHandler(SetPeriodic);
                tb.Items.Insert(ix++, btn);
            }
        }
        public void Initialize()
        {
            CreateNetwork();
            InitializeNeighborhood();
            Reset();
        }
        public Bitmap Reset()
        {
            InitializeMemory();
            _iterators.Clear();
            _iteratorIx = -1;
            _step = 0;
            _currentCell = 0;
            InitializeState();
            _colors.Clear();
            for (int ix = 0; ix < StateProvider.Count; ix++)
            {                
                _colors.Add(StateProvider.GetState(ix).StateColor.ToArgb());
            }
            MemoryStream ms = new MemoryStream(Resources.expressions);
            RuleTable rt = RuleTable.FromStream(ms);
            rt.Factory = new ExpressionFactory()
            {
                ValueProvider = this
            };
            for (int ix = 0; ix < StateProvider.Count; ix++)
            {
                StateProvider.GetState(ix).CompileTransitions(rt, StateProvider);
            }

            return GetImage();
        }
        public Bitmap Step()
        {
            Bitmap bmp = null;
            if (_step > 0)
            {
                for (_currentCell = 0; _currentCell < _width; _currentCell++)
                {
                    int stix = _cells[_currentCell].ActualState;
                    ICAState st = StateProvider.GetState(stix);
                    if (_optimizer != null)
                    {
                        int[] nh = _cells[_currentCell].GetNeighborhood();
                        for (int ix = 0; ix < nh.Length; ix++)
                        {
                            nh[ix] = _cells[nh[ix]].ActualState;
                        }
                        int nwst = _optimizer.GetTransition(stix, nh);
                        if (nwst >= 0)
                        {
                            _cells[_currentCell].NextState = nwst;
                        }
                        else
                        {
                            nwst = st.CheckTransitions();
                            if (nwst >= 0)
                            {
                                _cells[_currentCell].NextState = nwst;
                                _optimizer.AddKnownTransition(stix, nh, nwst);
                            }
                            else
                            {
                                _cells[_currentCell].NextState = _cells[_currentCell].ActualState;
                            }
                        }
                    }
                    else
                    {
                        int nwst = st.CheckTransitions();
                        if (nwst >= 0)
                        {
                            _cells[_currentCell].NextState = nwst;
                        }
                        else
                        {
                            _cells[_currentCell].NextState = _cells[_currentCell].ActualState;
                        }
                    }
                }                
            }
            bmp = GetImage();
            _memoryPos++;
            _memoryPos %= _history;
            _step++;
            return bmp;
        }
        public Bitmap ZoomIn()
        {
            if (_zoom < 32)
            {
                _zoom <<= 1;
            }
            return GetImage();
        }
        public Bitmap ZoomOut()
        {
            if (_zoom > 1)
            {
                _zoom >>= 1;
            }
            return GetImage();
        }
        #endregion
        private Bitmap GetImage()
        {
            int w = _width * _zoom;
            Bitmap bmp = new Bitmap(w, _history * _zoom);
            BitmapData bdata = null;
            try
            {
                for (int ix = 0; ix < _cells.Length; ix++)
                {
                    _memory[ix, _memoryPos] = _colors[_cells[ix].GetState(_step)];
                }
                bdata = bmp.LockBits(new Rectangle(0, 0, w, _history * _zoom), ImageLockMode.ReadWrite, PixelFormat.Format32bppRgb);
                int[] data = new int[w * _history * _zoom];
                int iy = ((_step < _history) || (_memoryPos == 99)) ? 0 : (_memoryPos + 1);
                int c = 0;
                int px = 0;
                while (c < Math.Min(_history, _step + 1))
                {                    
                    for (int ih = 0; ih < _zoom; ih++)
                    {
                        for (int ix = 0; ix < _width; ix++)
                        {
                            for (int iz = 0; iz < _zoom; iz++)
                            {
                                data[px++] = _memory[ix, iy];
                            }
                        }
                    }
                    iy++;
                    if (iy >= _history)
                    {
                        iy = 0;
                    }
                    c++;
                }
                Marshal.Copy(data, 0, bdata.Scan0, w * _history * _zoom);
                return bmp;
            }
            finally
            {
                if (bdata != null)
                {
                    bmp.UnlockBits(bdata);
                }
            }
        }
        private void CreateNetwork()
        {
            _cells = new ICACell[_width];
            if ((_initialState == null) || (_initialState.Length != _width))
            {
                _initialState = new int[_width];
            }
            for (int ix = 0; ix < _width; ix++)
            {
                _cells[ix] = _cellProvider.GetCell(_cbCells.SelectedItem.GetType(), 0);
            }
        }
        private void InitializeState()
        {
            for (int ix = 0; ix < _width; ix++)
            {
                _cells[ix].ActualState = _initialState[ix];
            }
        }
        private void InitializeMemory()
        {
            _memory = new int[_width, _history];
            _memoryPos = 0;
            for (int iy = 0; iy < _history; iy++)
            {
                for (int ix = 0; ix < _width; ix++)
                {
                    _memory[ix, iy] = Color.Black.ToArgb();
                }
            }
        }
        private void InitializeNeighborhood()
        {
            if ((_cells != null) && (_neighborhood != null))
            {
                for (int ix = 0; ix < _cells.Length; ix++)
                {
                    List<int> nc = new List<int>();
                    for (int n = 0; n < _neighborhood.Length; n++)
                    {
                        if (_neighborhood[n] != 0)
                        {
                            if (_periodic)
                            {
                                if (ix + _neighborhood[n] < 0)
                                {
                                    nc.Add(_width + _neighborhood[n]);
                                }
                                else if (ix + _neighborhood[n] >= _width)
                                {
                                    nc.Add(_neighborhood[n] - 1);
                                }
                                else
                                {
                                    nc.Add(ix + _neighborhood[n]);
                                }
                            }
                            else
                            {
                                if ((ix + _neighborhood[n] >= 0) && (ix + _neighborhood[n] < _width))
                                {
                                    nc.Add(ix + _neighborhood[n]);
                                }
                            }
                        }
                    }
                    _cells[ix].AddNeighbours(nc.ToArray());
                }
            }
        }
        private void EditorChange(object sender, EventArgs e)
        {
            _neighborhood = _editor.Neighborhood;
            InitializeNeighborhood();
        }
        private void WidthChanged(object sender, EventArgs e)
        {
            int w = 0;
            if (int.TryParse(_textWidth.Text, out w))
            {
                Width = w;
            }
            _textWidth.Text = _width.ToString();
        }
        private void HistoryChanged(object sender, EventArgs e)
        {
            int w = 0;
            if (int.TryParse(_textHistory.Text, out w))
            {
                History = w;
            }
            _textHistory.Text = _history.ToString();
        }
        private void SetPeriodic(object sender, EventArgs e)
        {
            ToolStripButton btn = sender as ToolStripButton;
            if (btn != null)
            {
                _periodic = btn.Checked;
                if (_periodic)
                {
                    btn.ToolTipText = UIResources.TTP_PERIODIC;
                }
                else
                {
                    btn.ToolTipText = UIResources.TTP_FINITE;
                }
            }
        }
        private void Initialize(object sender, EventArgs e)
        {
            GridEditor dlg = new GridEditor();
            dlg.StateProvider = StateProvider;
            dlg.ItemSize = new Size(_width, 1);
            if (_initialState != null)
            {
                dlg.SetImageRow(0, _initialState);
            }
            if (dlg.ShowDialog() == DialogResult.OK)
            {
                _initialState = dlg.GetImageRow(0);
                InitializeState();
            }
        }
    }
}
