﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Text;
using BNFUP.Rules;
using CommonObjects.Interfaces;

namespace CellularAutomata.Interfaces
{
    /// <summary>
    /// Cellular automata cell state
    /// </summary>
    public interface ICAState
    {
        /// <summary>
        /// Property values provider
        /// </summary>
        IPropertyProvider PropertyProvider { get; set; }
        /// <summary>
        /// Name of the property with the state color
        /// </summary>
        string Color { get; set; }
        /// <summary>
        /// Color of the state
        /// </summary>
        Color StateColor { get; }
        /// <summary>
        /// Gates or sets the state name
        /// </summary>
        string Name { get; set; }
        /// <summary>
        /// Enumeartes the properties of the state
        /// </summary>
        IEnumerable<IObjectProperty> Properties { get; }
        /// <summary>
        /// Enumeartes the names of the states to which this state has transitions
        /// </summary>
        IEnumerable<string> Transitions { get; }
        /// <summary>
        /// Gets the value of a property without type
        /// </summary>
        /// <param name="name">Name of the property</param>
        /// <returns>Property value with undefined type</returns>
        object GetValue(string name);
        /// <summary>
        /// Gets the value of a property as a variant value
        /// </summary>
        /// <param name="name">Name of the property</param>
        /// <returns>Property value as an IVariantValue interface</returns>
        IVariantValue GetVarValue(string name);
        /// <summary>
        /// Add a new property to the state
        /// </summary>
        /// <param name="name">Property name</param>
        void AddProperty(string name);
        /// <summary>
        /// Remove a property from the state
        /// </summary>
        /// <param name="name">Property name</param>
        void RemoveProperty(string name);
        /// <summary>
        /// Change a peorperty of the state
        /// </summary>
        /// <param name="name">Name of the property</param>
        /// <param name="p">New property to substitute</param>
        void ChangeProperty(string name, IObjectProperty p);
        /// <summary>
        /// Add a transtition expression to another state
        /// </summary>
        /// <param name="toname">Name of the final state</param>
        /// <param name="expression">Text with the expression that defines the transition conditions</param>
        void SetTransition(string toname, string expression);
        /// <summary>
        /// Gets the expression code of the transition to another state
        /// </summary>
        /// <param name="toname">Name of the final state</param>
        /// <returns></returns>
        string GetTransition(string toname);
        /// <summary>
        /// Compile the expressions of the transitions to other states
        /// </summary>
        /// <param name="rt">RuleTable object to compile the expressions</param>
        /// <param name="statep">State provider</param>
        void CompileTransitions(RuleTable rt, ICAStateProvider statep);
        /// <summary>
        /// Check if the state has to change to another one
        /// </summary>
        /// <returns>The new state index or -1 for no change</returns>
        int CheckTransitions();
    }
}
