﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using CommonObjects.Interfaces;
using CellularAutomata.Interfaces;

namespace CellularAutomata.States
{
    [Serializable]
    public class BasicCAStateProvider : ICAStateProvider
    {
        private IPropertyProvider _propertyProvider = null;
        private List<ICAState> _states = new List<ICAState>();
        private string _uid = Guid.NewGuid().ToString("N");
        private Type _defaultStateType = typeof(BasicState);
        public BasicCAStateProvider()
        {
        }
        #region ICAStateProvider
        public string UID
        {
            get
            {
                return _uid;
            }
        }
        public string Name { get; set; }
        public int Count
        {
            get
            {
                return _states.Count;
            }
        }
        public IPropertyProvider PropertyProvider
        {
            get
            {
                return _propertyProvider;
            }
            set
            {
                _propertyProvider = value;
                foreach (ICAState s in _states)
                {
                    s.PropertyProvider = value;
                }
            }
        }
        public IEnumerable<ICAState> States
        {
            get
            {
                foreach (ICAState s in _states)
                {
                    yield return s;
                }
            }
        }
        public Type DefaultStateType
        {
            get
            {
                return _defaultStateType;
            }
            set
            {
                if (value.GetInterface("ICAState") != null)
                {
                    _defaultStateType = value;
                }
            }
        }
        public ICAState GetState(string name)
        {
            foreach (ICAState s in _states)
            {
                if (string.Compare(s.Name, name, true) == 0)
                {
                    return s;
                }
            }
            return null;
        }
        public ICAState GetState(int index)
        {
            return _states[index];
        }
        public int IndexOf(ICAState s)
        {
            return _states.IndexOf(s);
        }
        public int IndexOf(string name)
        {
            for(int ix = 0; ix < _states.Count; ix++)
            {
                if (string.Compare(_states[ix].Name, name, true) == 0)
                {
                    return ix;
                }
            }
            return -1;
        }
        public bool ExistsState(string name)
        {
            foreach (ICAState s in _states)
            {
                if (string.Compare(s.Name, name, true) == 0)
                {
                    return true;
                }
            }
            return false;
        }
        public void AddState(ICAState s)
        {
            if (!_states.Contains(s))
            {
                _states.Add(s);
                s.PropertyProvider = _propertyProvider;
            }
        }
        public void AddStates(IEnumerable<ICAState> ss)
        {
            foreach (ICAState s in ss)
            {
                AddState(s);
            }
        }
        public void RemoveState(string name)
        {
            ICAState s = GetState(name);
            if (s != null)
            {
                _states.Remove(s);
            }
        }
        public void RemoveState(ICAState s)
        {
            _states.Remove(s);
        }
        public void ChangeState(string name, ICAState s)
        {
            ICAState ss = GetState(name);
            if (ss != null)
            {
                if (string.Compare(s.Name, name, true) == 0)
                {
                    int ix = _states.IndexOf(ss);
                    _states[ix] = s;
                    s.PropertyProvider = _propertyProvider;
                }
                else
                {
                    _states.Remove(ss);
                    AddState(s);
                }
            }
        }
        public ICAState CreateState(string name)
        {
            ICAState s = (ICAState)_defaultStateType.GetConstructor(Type.EmptyTypes).Invoke(null);
            s.Name = name;
            return s;
        }
        #endregion
    }
}
