﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Drawing;
using BNFUP.Rules;
using BNFUP.Interfaces;
using CommonObjects.Interfaces;
using CommonObjects.Expressions;
using CellularAutomata.Interfaces;
using CellularAutomata.Data;
using CellularAutomata.Properties;
using CellularAutomata.Expressions;

namespace CellularAutomata.States
{
    [Serializable]
    public class BasicState : ICAState
    {
        private IPropertyProvider _provider = null;
        private string _color = "";
        private List<PropertyValue> _properties = new List<PropertyValue>();
        private Dictionary<string, string> _transitions = new Dictionary<string, string>();
        [NonSerialized]
        private Dictionary<int, ExprBase> _triggers = new Dictionary<int, ExprBase>();
        public BasicState()
        {
        }
        #region ICAState
        public IPropertyProvider PropertyProvider
        {
            get
            {
                return _provider;
            }
            set
            {
                foreach (PropertyValue pv in _properties)
                {
                    if (!value.ExistsProperty(pv.Name))
                    {
                        throw new Exception(string.Format(UIResources.ERR_UNDEFCONST, Name, pv.Name));
                    }
                }
                _provider = value;
            }
        }
        public string Color
        {
            get
            {
                return _color;
            }
            set
            {
                if (_provider != null)
                {
                    IObjectProperty p = _provider.GetProperty(value);
                    if ((p != null) && (p.Value.NativeType == typeof(Color)))
                    {
                        _color = value;
                        AddProperty(value);
                    }
                }
            }
        }
        public Color StateColor
        {
            get
            {
                if (_provider != null)
                {
                    IVariantValue p = GetVarValue(_color);
                    if (p != null)
                    {
                        return p.AsColor;
                    }
                }
                return System.Drawing.Color.Transparent;
            }
        }
        public string Name { get; set; }
        public IEnumerable<IObjectProperty> Properties
        {
            get
            {
                foreach (PropertyValue p in _properties)
                {
                    yield return _provider.GetProperty(p.Name);
                }
            }
        }
        public IEnumerable<string> Transitions
        {
            get
            {
                foreach (string k in _transitions.Keys)
                {
                    yield return k;
                }
            }
        }
        public object GetValue(string name)
        {
            foreach (PropertyValue pv in _properties)
            {
                if (string.Compare(pv.Name, name, true) == 0)
                {
                    return pv;
                }
            }
            return null;
        }
        public IVariantValue GetVarValue(string name)
        {
            foreach (PropertyValue pv in _properties)
            {
                if (string.Compare(pv.Name, name, true) == 0)
                {
                    return pv.VarValue;
                }
            }
            return null;
        }
        public void AddProperty(string name)
        {
            IObjectProperty p = _provider.GetProperty(name);
            if (p != null)
            {
                for (int ix = 0; ix < _properties.Count; ix++)
                {
                    if (string.Compare(_properties[ix].Name, name, true) == 0)
                    {
                        return;
                    }
                }
                _properties.Add(PropertyValue.FromProperty(p));
            }
        }
        public void RemoveProperty(string name)
        {
            for (int ix = _properties.Count - 1; ix >= 0; ix--)
            {
                if (string.Compare(_properties[ix].Name, name, true) == 0)
                {
                    _properties.RemoveAt(ix);
                    break;
                }
            }
        }
        public void ChangeProperty(string name, IObjectProperty p)
        {
            foreach (PropertyValue pv in _properties)
            {
                if (string.Compare(pv.Name, name, true) == 0)
                {
                    pv.ChangeName(p.Name);
                    if (p.Value.NativeType != pv.VarValue.NativeType)
                    {
                        pv.VarValue = pv.VarValue.ConvertFrom(p.Value);
                    }
                    return;
                }
            }
        }
        public void SetTransition(string toname, string expression)
        {
            if (string.IsNullOrEmpty(expression))
            {
                _transitions.Remove(toname);
            }
            else
            {
                _transitions[toname] = expression;
            }
        }
        public string GetTransition(string toname)
        {
            if (_transitions.ContainsKey(toname))
            {
                return _transitions[toname];
            }
            return "";
        }
        public void CompileTransitions(RuleTable rt, ICAStateProvider statep)
        {
            _triggers = new Dictionary<int, ExprBase>();
            foreach (string k in _transitions.Keys)
            {
                StringReader sr = new StringReader(_transitions[k]);
                ExprBase expr = rt.Build(sr) as ExprBase;
                _triggers[statep.IndexOf(k)] = expr;
            }
        }
        public int CheckTransitions()
        {
            foreach (int k in _triggers.Keys)
            {
                if (_triggers[k].AsBool == true)
                {
                    return k;
                }
            }
            return -1;
        }
        #endregion
        public override string ToString()
        {
            return Name;
        }
    }
}
