﻿using System;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using CommonObjects.Interfaces;
using CommonObjects.Properties;

namespace CommonObjects.Data
{
    [Serializable]
    public class BasicPropertyProvider : IPropertyProvider
    {
        private List<IObjectProperty> _properties = new List<IObjectProperty>();
        private string _uid = Guid.NewGuid().ToString("N");
        private Type _defaultPropertyType;

        public BasicPropertyProvider()
        {
            Name = UIResources.TXT_NONAME;
        }
        #region IPropertyProvider
        public string UID
        {
            get
            {
                return _uid;
            }
        }
        public string Name { get; set; }
        public IEnumerable<IObjectProperty> Properties
        {
            get
            {
                foreach (IObjectProperty p in _properties)
                {
                    yield return p;
                }
            }
        }
        public Type DefaultPropertyType
        {
            get
            {
                return _defaultPropertyType;
            }
            set
            {
                if (value.GetInterface("IObjectProperty") == null)
                {
                    throw new Exception(UIResources.ERR_IVARIANTVALUE);
                }
                _defaultPropertyType = value;
            }
        }
        public IObjectProperty GetProperty(string name)
        {
            foreach (IObjectProperty p in _properties)
            {
                if (string.Compare(p.Name, name, true) == 0)
                {
                    return p;
                }
            }
            return null;
        }
        public IObjectProperty GetProperty(int index)
        {
            return _properties[index];
        }
        public int IndexOf(IObjectProperty p)
        {
            return _properties.IndexOf(p);
        }
        public int IndexOf(string name)
        {
            for (int ix = 0; ix < _properties.Count; ix++)
            {
                if (string.Compare(_properties[ix].Name, name, true) == 0)
                {
                    return ix;
                }
            }
            return -1;
        }
        public bool ExistsProperty(string name)
        {
            foreach (IObjectProperty p in _properties)
            {
                if (string.Compare(p.Name, name, true) == 0)
                {
                    return true;
                }
            }
            return false;
        }
        public void AddProperty(IObjectProperty p)
        {
            if (!_properties.Contains(p))
            {
                Regex reg = new Regex("^[A-Za-z][_0-9A-Za-z]*$");
                if (reg.IsMatch(p.Name))
                {
                    _properties.Add(p);
                }
                else
                {
                    throw new Exception(UIResources.ERR_BADNAME);
                }
            }
        }
        public void AddProperties(IEnumerable<IObjectProperty> ps)
        {
            foreach (IObjectProperty p in ps)
            {
                AddProperty(p);
            }
        }
        public void RemoveProperty(string name)
        {
            IObjectProperty p = GetProperty(name);
            if (p != null)
            {
                _properties.Remove(p);
            }
        }
        public void RemoveProperty(IObjectProperty p)
        {
            _properties.Remove(p);
        }
        public void ChangeProperty(string name, IObjectProperty p)
        {
            IObjectProperty pp = GetProperty(name);
            if (pp != null)
            {
                if (string.Compare(p.Name, name, true) == 0)
                {
                    int ix = _properties.IndexOf(pp);
                    _properties[ix] = p;
                }
                else
                {
                    _properties.Remove(pp);
                    AddProperty(p);
                }
            }
        }
        public IObjectProperty CreateProperty(string name, object value)
        {
            IObjectProperty p = (IObjectProperty)_defaultPropertyType.GetConstructor(new Type[] { typeof(object) }).Invoke(new object[] { value });
            p.Name = name;
            return p;
        }
        #endregion
    }
}
