﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using CommonObjects.Interfaces;
using System.Drawing;

namespace CommonObjects.Data
{
    [Serializable]
    public abstract class VariantValue : IVariantValue
    {
        protected bool _mutable = true;
        public VariantValue()
        {
        }
        #region IVariantValue
        public bool Mutable
        {
            get
            {
                return _mutable;
            }
        }
        public abstract int AsInt { get; set; }
        public abstract double AsDouble { get; set; }
        public abstract Color AsColor { get; set; }
        public abstract string AsString { get; set; }
        public abstract bool AsBool { get; set; }
        public abstract object AsObject { get; set; }
        public abstract Type NativeType { get; }
        public IVariantValue FromObject(object value)
        {
            if (value.GetType() == typeof(int))
            {
                return new IntVariantValue((int)value);
            }
            if (value.GetType() == typeof(double))
            {
                return new DoubleVariantValue((double)value);
            }
            if (value.GetType() == typeof(Color))
            {
                return new ColorVariantValue((Color)value);
            }
            if (value.GetType() == typeof(bool))
            {
                return new BoolVariantValue((bool)value);
            }
            if (value.GetType() == typeof(string))
            {
                return new StringVariantValue((string)value);
            }
            return null;
        }
        public abstract IVariantValue ConvertFrom(IVariantValue value);
        public abstract IVariantValue Clone();
        #endregion
    }
    [Serializable]
    public class IntVariantValue : VariantValue
    {
        private int _value;
        public IntVariantValue()
        {
        }
        public IntVariantValue(int value)
        {
            _value = value;
        }
        #region IVariantValue
        public override int AsInt
        {
            get
            {
                return _value;
            }
            set
            {
                _value = value;
            }
        }
        public override double AsDouble
        {
            get
            {
                return (double)_value;
            }
            set
            {
                _value = (int)value;
            }
        }
        public override Color AsColor
        {
            get
            {
                return Color.FromArgb(_value);
            }
            set
            {
                _value = value.ToArgb();
            }
        }
        public override string AsString
        {
            get
            {
                return _value.ToString();
            }
            set
            {
                int.TryParse(value, out _value);
            }
        }
        public override bool AsBool
        {
            get
            {
                return _value != 0;
            }
            set
            {
                _value = value ? 1 : 0;
            }
        }
        public override object AsObject
        {
            get
            {
                return _value;
            }
            set
            {
                _value = Convert.ToInt32(value);
            }
        }
        public override Type NativeType
        {
            get
            {
                return typeof(int);
            }
        }
        public override IVariantValue ConvertFrom(IVariantValue value)
        {
            return new IntVariantValue(value.AsInt);
        }
        public override IVariantValue Clone()
        {
            return new IntVariantValue(_value);
        }
        #endregion
    }
    [Serializable]
    public class ColorVariantValue : VariantValue
    {
        private int _value;
        public ColorVariantValue()
        {
        }
        public ColorVariantValue(Color value)
        {
            AsColor = value;
        }
        #region IVariantValue
        public override int AsInt
        {
            get
            {
                return _value;
            }
            set
            {
                _value = value;
            }
        }
        public override double AsDouble
        {
            get
            {
                return (double)_value;
            }
            set
            {
                _value = (int)value;
            }
        }
        public override Color AsColor
        {
            get
            {
                return Color.FromArgb(_value);
            }
            set
            {
                _value = value.ToArgb();
            }
        }
        public override string AsString
        {
            get
            {
                return Color.FromArgb(_value).ToString();
            }
            set
            {
                _value = Color.FromName(value).ToArgb();
            }
        }
        public override bool AsBool
        {
            get
            {
                return _value != Color.Black.ToArgb();
            }
            set
            {
                _value = value ? Color.White.ToArgb() : Color.Black.ToArgb();
            }
        }
        public override object AsObject
        {
            get
            {
                return _value;
            }
            set
            {
                _value = ((Color)value).ToArgb();
            }
        }
        public override Type NativeType
        {
            get
            {
                return typeof(Color);
            }
        }
        public override IVariantValue ConvertFrom(IVariantValue value)
        {
            return new ColorVariantValue(value.AsColor);
        }
        public override IVariantValue Clone()
        {
            return new ColorVariantValue(Color.FromArgb(_value));
        }
        #endregion
    }
    [Serializable]
    public class DoubleVariantValue : VariantValue
    {
        private double _value;
        public DoubleVariantValue()
        {
        }
        public DoubleVariantValue(double value)
        {
            _value = value;
        }
        #region IVariantValue
        public override int AsInt
        {
            get
            {
                return (int)_value;
            }
            set
            {
                _value = value;
            }
        }
        public override double AsDouble
        {
            get
            {
                return _value;
            }
            set
            {
                _value = value;
            }
        }
        public override Color AsColor
        {
            get
            {
                return Color.FromArgb((int)_value);
            }
            set
            {
                _value = value.ToArgb();
            }
        }
        public override string AsString
        {
            get
            {
                return _value.ToString();
            }
            set
            {
                double.TryParse(value, out _value);
            }
        }
        public override bool AsBool
        {
            get
            {
                return _value != 0;
            }
            set
            {
                _value = value ? 1 : 0;
            }
        }
        public override object AsObject
        {
            get
            {
                return _value;
            }
            set
            {
                _value = Convert.ToDouble(value);
            }
        }
        public override Type NativeType
        {
            get
            {
                return typeof(double);
            }
        }
        public override IVariantValue ConvertFrom(IVariantValue value)
        {
            return new DoubleVariantValue(value.AsDouble);
        }
        public override IVariantValue Clone()
        {
            return new DoubleVariantValue(_value);
        }
        #endregion
    }
    [Serializable]
    public class StringVariantValue : VariantValue
    {
        private string _value;
        public StringVariantValue()
        {
        }
        public StringVariantValue(string value)
        {
            _value = value;
        }
        #region IVariantValue
        public override int AsInt
        {
            get
            {
                int v = 0;
                int.TryParse(_value, out v);
                return v;
            }
            set
            {
                _value = value.ToString();
            }
        }
        public override double AsDouble
        {
            get
            {
                double v = double.NaN;
                double.TryParse(_value, out v);
                return v;
            }
            set
            {
                _value = value.ToString();
            }
        }
        public override Color AsColor
        {
            get
            {
                return Color.FromArgb(AsInt);
            }
            set
            {
                _value = value.ToArgb().ToString();
            }
        }
        public override string AsString
        {
            get
            {
                return _value;
            }
            set
            {
                _value = value;
            }
        }
        public override bool AsBool
        {
            get
            {
                return string.IsNullOrEmpty(_value) || AsInt == 0;
            }
            set
            {
                _value = value ? "1" : "0";
            }
        }
        public override object AsObject
        {
            get
            {
                return _value;
            }
            set
            {
                _value = value as string;
            }
        }
        public override Type NativeType
        {
            get
            {
                return typeof(string);
            }
        }
        public override IVariantValue ConvertFrom(IVariantValue value)
        {
            return new StringVariantValue(value.AsString);
        }
        public override IVariantValue Clone()
        {
            return new StringVariantValue(_value);
        }
        #endregion
    }
    [Serializable]
    public class BoolVariantValue : VariantValue
    {
        private bool _value;
        public BoolVariantValue()
        {
        }
        public BoolVariantValue(bool value)
        {
            _value = value;
        }
        #region IVariantValue
        public override int AsInt
        {
            get
            {
                return _value ? 1 : 0;
            }
            set
            {
                _value = value != 0;
            }
        }
        public override double AsDouble
        {
            get
            {
                return _value ? 1 : 0;
            }
            set
            {
                _value = value != 0;
            }
        }
        public override Color AsColor
        {
            get
            {
                return _value ? Color.White : Color.Black;
            }
            set
            {
                _value = value.ToArgb() != Color.Black.ToArgb();
            }
        }
        public override string AsString
        {
            get
            {
                return _value.ToString();
            }
            set
            {
                _value = Convert.ToBoolean(value);
            }
        }
        public override bool AsBool
        {
            get
            {
                return _value;
            }
            set
            {
                _value = value;
            }
        }
        public override object AsObject
        {
            get
            {
                return _value;
            }
            set
            {
                _value = Convert.ToBoolean(value);
            }
        }
        public override Type NativeType
        {
            get
            {
                return typeof(bool);
            }
        }
        public override IVariantValue ConvertFrom(IVariantValue value)
        {
            return new BoolVariantValue(value.AsBool);
        }
        public override IVariantValue Clone()
        {
            return new BoolVariantValue(_value);
        }
        #endregion
    }
}
