﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using CommonObjects.Data;

namespace CommonObjects.Interfaces
{
    /// <summary>
    /// Property manager
    /// </summary>
    public interface IPropertyProvider
    {
        /// <summary>
        /// Gets the unique identifier of the provider
        /// </summary>
        string UID { get; }
        /// <summary>
        /// Gets or sets the name of the provider
        /// </summary>
        string Name { get; set; }
        /// <summary>
        /// Enumerates all the properties
        /// </summary>
        IEnumerable<IObjectProperty> Properties { get; }
        /// <summary>
        /// Gets or sets the default type to create new properties. The type must implement the IObjectProperty interface
        /// </summary>
        Type DefaultPropertyType { get; set; }
        /// <summary>
        /// Get a property by his name
        /// </summary>
        /// <param name="name">Name of the property</param>
        /// <returns>Property with the given name</returns>
        IObjectProperty GetProperty(string name);
        /// <summary>
        /// Get a property by index
        /// </summary>
        /// <param name="index">Index of the property</param>
        /// <returns>Property in the given index</returns>
        IObjectProperty GetProperty(int index);
        /// <summary>
        /// Gets the index of a property
        /// </summary>
        /// <param name="p">Property to get the index</param>
        /// <returns>The base 0 index of the property, or -1 if doesn't exists</returns>
        int IndexOf(IObjectProperty p);
        /// <summary>
        /// Gets the index of a property by his name
        /// </summary>
        /// <param name="name">Name of the property</param>
        /// <returns>The base 0 index of the property, or -1 if doesn't exists</returns>
        int IndexOf(string name);
        /// <summary>
        /// Test if the provider owns a property
        /// </summary>
        /// <param name="name">Name of the property</param>
        /// <returns>true if there is a property with the given name</returns>
        bool ExistsProperty(string name);
        /// <summary>
        /// Add a new property to the collection
        /// </summary>
        /// <param name="cons">New property to add</param>
        void AddProperty(IObjectProperty cons);
        /// <summary>
        /// Add a collection of new properties
        /// </summary>
        /// <param name="ps">Property enumerator</param>
        void AddProperties(IEnumerable<IObjectProperty> ps);
        /// <summary>
        /// Remove a property from the collection
        /// </summary>
        /// <param name="name">Property name</param>
        void RemoveProperty(string name);
        /// <summary>
        /// Remove a propewrty from the collection
        /// </summary>
        /// <param name="p">Property to remove</param>
        void RemoveProperty(IObjectProperty p);
        /// <summary>
        /// Change a property of the collection
        /// </summary>
        /// <param name="name">Name of the property to change</param>
        /// <param name="p">New property to substitute</param>
        void ChangeProperty(string name, IObjectProperty p);
        /// <summary>
        /// Creates a new property with the default type in DefaultPropertyType
        /// </summary>
        /// <param name="name">New property name</param>
        /// <param name="value">New property value</param>
        /// <returns>The new property created</returns>
        IObjectProperty CreateProperty(string name, object value);        
    }
}
