﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Configuration;
using System.Runtime.Serialization.Formatters.Binary;
using System.IO;
using System.Windows.Forms;
using CommonObjects.Interfaces;
using CellularAutomata.Interfaces;
using CellularAutomata.States;
using WinCA.Properties;
using WinCA.Interfaces;
using WinCA.Utils;

namespace WinCA.Forms
{
    public partial class frmStates : Form, ICAStateManager
    {
        private IMainForm _mainForm = null;
        private string _fileName = "";
        private ICAStateProvider _provider = null;
        private PropertySerializationContainer _properties = null;
        private ICAState _actualState = null;
        private ITransitionManager _transMgr = null;
        private ICAManager _automata = null;

        public frmStates()
        {
            InitializeComponent();
            Translate();
            Type tp = Type.GetType(ConfigurationManager.AppSettings["defaultStateProviderType"]);
            Provider = tp.GetConstructor(Type.EmptyTypes).Invoke(null) as ICAStateProvider;
        }
        public frmStates(StateSerializationContainer ls, string filename)
        {
            InitializeComponent();
            Provider = ls.StateProvider;
            _properties = ls.Properties;
            _fileName = filename;
            Translate();
        }
        #region IStateManager
        public event CAStateChangeEventHandler StateChanged = null;
        public ICAStateProvider Provider
        {
            get
            {
                return _provider;
            }
            set
            {
                _provider = value;
                GetStates();
            }
        }
        public IPropertyProvider VariableProvider
        {
            get
            {
                if (_properties != null)
                {
                    return _properties.VariableProvider;
                }
                return null;
            }
        }
        public IMainForm MainForm
        {
            get
            {
                return _mainForm;
            }
            set
            {
                _mainForm = value;
                _mainForm.OnPropertyManagerChanged += new PropertyManagerChangedEventHandler(PropertyManagerChanged);
                cbProperties.Items.Clear();
                foreach (IPropertyManager pm in _mainForm.PropertyManagers)
                {
                    cbProperties.Items.Add(pm);
                }
                if (_properties != null)
                {
                    if (!SelectPropertyManager(_properties.ConstantProvider.UID))
                    {
                        _mainForm.CreatePropertyManager(_properties, _properties.ConstantProvider.Name);
                        SelectPropertyManager(_properties.ConstantProvider.UID);
                    }
                }
                else
                {
                    if (cbProperties.Items.Count == 0)
                    {
                        _mainForm.CreatePropertyManager(null, "");
                    }
                    cbProperties.SelectedIndex = 0;
                }
            }
        }
        #endregion
        public override string ToString()
        {
            return _provider.Name;
        }
        protected override void OnFormClosed(FormClosedEventArgs e)
        {
            base.OnFormClosed(e);
            if ((_transMgr != null) && (_transMgr is Form))
            {
                (_transMgr as Form).Close();
            }
        }
        private string NewName
        {
            get
            {
                string name = UIResources.TXT_NONAME;
                int n = 1;
                while (TestName(name + n.ToString()) > 0)
                {
                    n++;
                }
                return name + n.ToString();
            }
        }
        private int TestName(string name)
        {
            int nc = 0;
            foreach (ICAState p in lbStates.Items)
            {
                if (string.Compare(p.Name, name, true) == 0)
                {
                    nc++;
                }
            }
            return nc;
        }
        private void Translate()
        {
            Text = string.Format(UIResources.TTL_STATES, _fileName);
            opFile.Text = UIResources.MNU_FILE;
            opSave.Text = UIResources.OPT_SAVESTS;
            opSaveAs.Text = UIResources.OPT_SAVESTSAS;
            lProperties.Text = UIResources.LAB_PROPMANAGER;
            lProps.Text = UIResources.LAB_PROPERTIES;
            bSelect.ToolTipText = UIResources.TTP_SELECTPROPS;
            lStProps.Text = UIResources.LAB_STPROPS;
            bRemove.ToolTipText = UIResources.TTP_REMOVEPROPS;
            lStates.Text = UIResources.LAB_STATES;
            bAdd.ToolTipText = UIResources.TTP_ADDSTATE;
            bDelete.ToolTipText = UIResources.TTP_DELETESTATE;
            lName.Text = UIResources.LAB_NAME;
            lColor.Text = UIResources.LAB_COLOR;
            bChange.Text = UIResources.BTN_CHANGE;
            bTransitions.ToolTipText = UIResources.TTP_SHOWTRANSITIONS;
            bShowCA.ToolTipText = UIResources.TTP_SHOWCA;
        }
        private void OpenFile(string filename)
        {
            FileStream f = null;
            try
            {
                f = new FileStream(filename, FileMode.Open, FileAccess.Read, FileShare.Read);
                BinaryFormatter fmt = new BinaryFormatter();
                StateSerializationContainer ls = fmt.Deserialize(f) as StateSerializationContainer;
                _properties = ls.Properties;
                Provider = ls.StateProvider;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
            finally
            {
                if (f != null)
                {
                    f.Close();
                }
            }
        }
        private void SaveFile(string filename)
        {
            FileStream f = null;
            try
            {
                f = new FileStream(filename, FileMode.OpenOrCreate, FileAccess.Write, FileShare.None);
                BinaryFormatter fmt = new BinaryFormatter();
                StateSerializationContainer ls = new StateSerializationContainer();
                ls.StateProvider = _provider;
                _provider.Name = filename;
                ls.Properties = _properties;
                _properties.VariableProvider.Name = filename;
                _properties.ConstantProvider.Name = filename;
                fmt.Serialize(f, ls);
                _fileName = filename;
                Text = string.Format(UIResources.TTL_STATES, _fileName);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
            finally
            {
                if (f != null)
                {
                    f.Close();
                }
            }
        }
        private void SaveCurrent()
        {
            if (_actualState != null)
            {
                if (string.IsNullOrEmpty(txtName.Text))
                {
                    throw new Exception(UIResources.ERR_EMPTYNAME);
                }
                if (string.Compare(txtName.Text, _actualState.Name, true) != 0)
                {
                    if (TestName(txtName.Text) > 0)
                    {
                        throw new Exception(UIResources.ERR_DUPLICATENAME);
                    }
                    else
                    {
                        _actualState.Name = txtName.Text;
                        lbStates.Items[lbStates.Items.IndexOf(_actualState)] = _actualState;
                    }
                }
                IObjectProperty pc = cbColor.SelectedItem as IObjectProperty;
                if (pc == null)
                {
                    throw new Exception(UIResources.ERR_COLORMANDATORY);
                }
                _actualState.Color = pc.Name;
            }
        }
        private bool SelectPropertyManager(string uid)
        {
            for (int ix = 0; ix < cbProperties.Items.Count; ix++)
            {
                IPropertyManager pm = cbProperties.Items[ix] as IPropertyManager;
                if (pm.ConstantProvider.UID == uid)
                {
                    cbProperties.SelectedIndex = ix;
                    return true;
                }
            }
            cbProperties.SelectedIndex = -1;
            return false;
        }
        private void PropertyManagerChanged(object sender, PropertyManagerEventArgs e)
        {
            switch (e.ChangeType)
            {
                case PropertyManagerEventArgs.Change.Added:
                    e.PropertyManager.PropertyChanged += new PropertyChangeEventHandler(PropertyChange);
                    cbProperties.Items.Add(e.PropertyManager);
                    break;
                case PropertyManagerEventArgs.Change.Removed:
                    e.PropertyManager.PropertyChanged -= new PropertyChangeEventHandler(PropertyChange);
                    cbProperties.Items.Remove(e.PropertyManager);
                    SelectPropertyManager(_properties.ConstantProvider.UID);
                    break;
            }            
        }
        private void PropertyChange(object sender, PropertyChangeEventArgs e)
        {
            if (e.Provider == _properties.ConstantProvider)
            {
                switch (e.ChangeType)
                {
                    case PropertyChangeEventArgs.Change.Added:
                        lbProps.Items.Add(e.Property);
                        if (e.Property.Value.NativeType == typeof(Color))
                        {
                            cbColor.Items.Add(e.Property);
                        }
                        break;
                    case PropertyChangeEventArgs.Change.Deleted:
                        lbProps.Items.Remove(e.Property);
                        lbStProps.Items.Remove(e.Property);
                        if (e.Property.Value.NativeType == typeof(Color))
                        {
                            cbColor.Items.Remove(e.Property);
                        }
                        break;
                    case PropertyChangeEventArgs.Change.NameChanged:
                    case PropertyChangeEventArgs.Change.TypeChanged:                        
                        foreach (ICAState st in _provider.States)
                        {
                            st.ChangeProperty(e.OldName, e.Property);
                        }
                        for (int ix = 0; ix < lbProps.Items.Count; ix++)
                        {
                            if (lbProps.Items[ix] == e.Property)
                            {
                                lbProps.Items[ix] = e.Property;
                            }
                        }
                        for (int ix = 0; ix < lbStProps.Items.Count; ix++)
                        {
                            if (lbStProps.Items[ix] == e.Property)
                            {
                                lbStProps.Items[ix] = e.Property;
                            }
                        }
                        break;
                }
            }
        }
        private void GetStates()
        {
            lbStates.Items.Clear();
            if (_provider != null)
            {
                foreach (ICAState s in _provider.States)
                {
                    lbStates.Items.Add(s);
                }
            }
        }
        private void Getproperties()
        {
            lbProps.Items.Clear();
            if (_properties != null)
            {
                foreach (IObjectProperty p in _properties.ConstantProvider.Properties)
                {
                    lbProps.Items.Add(p);
                }
            }
        }
        private void GetStateProperties()
        {
            lbStProps.Items.Clear();
            if (_actualState != null)
            {
                foreach (IObjectProperty p in _actualState.Properties)
                {
                    lbStProps.Items.Add(p);
                }
            }
        }

        private void cbProperties_SelectedIndexChanged(object sender, EventArgs e)
        {
            IPropertyManager pm = cbProperties.SelectedItem as IPropertyManager;            
            if (pm != null)
            {
                IPropertyProvider oldpc = pm.ConstantProvider;
                IPropertyProvider oldpv = pm.VariableProvider;
                try
                {
                    if (_properties == null)
                    {
                        _properties = new PropertySerializationContainer()
                        {
                            ConstantProvider = pm.ConstantProvider,
                            VariableProvider = pm.VariableProvider
                        };
                    }
                    else if (_properties.ConstantProvider.UID != pm.ConstantProvider.UID)
                    {
                        _properties.ConstantProvider = pm.ConstantProvider;
                        _properties.VariableProvider = pm.VariableProvider;
                    }
                    _provider.PropertyProvider = pm.ConstantProvider;
                    cbColor.Items.Clear();
                    foreach (IObjectProperty p in pm.ConstantProvider.Properties)
                    {
                        if (p.Value.NativeType == typeof(Color))
                        {
                            cbColor.Items.Add(p);
                        }
                    }
                    Getproperties();
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message);
                    _provider.PropertyProvider = oldpc;
                    _properties.ConstantProvider = oldpc;
                    _properties.VariableProvider = oldpv;
                    SelectPropertyManager(oldpc.UID);
                }
            }
        }

        private void bPropertyEditor_Click(object sender, EventArgs e)
        {
            try
            {
                _mainForm.CreatePropertyManager(_properties, _properties.ConstantProvider.Name);
                SelectPropertyManager(_properties.ConstantProvider.UID);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        private void bAdd_Click(object sender, EventArgs e)
        {
            try
            {
                ICAState sts = _provider.CreateState(NewName);
                lbStates.Items.Add(sts);
                lbStates.SelectedItem = sts;
                _provider.AddState(sts);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        private void lbStates_SelectedIndexChanged(object sender, EventArgs e)
        {
            ICAState sts = lbStates.SelectedItem as ICAState;
            if (sts != null)
            {
                try
                {
                    SaveCurrent();
                    txtName.Text = sts.Name;
                    IObjectProperty opc = _properties.ConstantProvider.GetProperty(sts.Color);
                    cbColor.SelectedItem = opc;
                    _actualState = sts;
                    GetStateProperties();
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message);
                }
            }
        }

        private void opSave_Click(object sender, EventArgs e)
        {
            if (string.IsNullOrEmpty(_fileName))
            {
                opSaveAs_Click(sender, e);
                return;
            }
            SaveFile(_fileName);
        }

        private void opSaveAs_Click(object sender, EventArgs e)
        {
            sfDlg.Filter = UIResources.FIL_STATES;
            if (sfDlg.ShowDialog() == DialogResult.OK)
            {
                SaveFile(sfDlg.FileName);
            }
        }

        private void bChange_Click(object sender, EventArgs e)
        {
            try
            {
                SaveCurrent();
                GetStateProperties();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        private void bSelect_Click(object sender, EventArgs e)
        {
            if (_actualState != null)
            {
                try
                {
                    foreach (IObjectProperty p in lbProps.SelectedItems)
                    {
                        _actualState.AddProperty(p.Name);
                        if (!lbStProps.Items.Contains(p))
                        {
                            lbStProps.Items.Add(p);
                        }
                    }
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message);
                }
            }
        }

        private void bRemove_Click(object sender, EventArgs e)
        {
            if (_actualState != null)
            {
                try
                {
                    foreach (IObjectProperty p in lbStProps.SelectedItems)
                    {
                        _actualState.RemoveProperty(p.Name);
                        lbStProps.Items.Remove(p);
                    }
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message);
                }
            }
        }

        private void lbStProps_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (_actualState != null)
            {
                try
                {
                    IObjectProperty p = lbStProps.SelectedItem as IObjectProperty;
                    if (p != null)
                    {
                        pgValue.SelectedObject = _actualState.GetValue(p.Name);
                    }
                    else
                    {
                        pgValue.SelectedObject = null;
                    }
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message);
                }
            }
        }

        private void bTransitions_Click(object sender, EventArgs e)
        {
            try
            {
                if (_provider.Count > 0)
                {
                    _transMgr = _mainForm.CreateTransitionManager(this, _fileName);
                    _transMgr.AutomataManager = _automata;
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        private void bShowCA_Click(object sender, EventArgs e)
        {
            try
            {
                if ((_automata != null) && (_automata is Form))
                {
                    (_automata as Form).FormClosed -= new FormClosedEventHandler(AutomataClosed);
                }
                _automata = _mainForm.CreateCAManger(this);
                if (_transMgr != null)
                {
                    _transMgr.AutomataManager = _automata;
                }
                if (_automata is Form)
                {
                    (_automata as Form).FormClosed += new FormClosedEventHandler(AutomataClosed);
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }
        private void AutomataClosed(object sender, FormClosedEventArgs e)
        {
            _automata = null;
            if (_transMgr != null)
            {
                _transMgr.AutomataManager = null;
            }
        }

        private void frmStates_FormClosed(object sender, FormClosedEventArgs e)
        {
            _mainForm.OnPropertyManagerChanged -= new PropertyManagerChangedEventHandler(PropertyManagerChanged);
        }

        private void frmStates_Load(object sender, EventArgs e)
        {
            Icon = Icon.Clone() as Icon;
            if ((_provider.Count > 0) && (_mainForm != null))
            {
                _transMgr = _mainForm.CreateTransitionManager(this, _fileName);
            }
        }
    }
}
