﻿using System.Collections.Generic;
using System.Drawing;
using System.ComponentModel;
using CommonObjects.Interfaces;

namespace WinCA.Utils
{
    public enum DataType
    {
        Int, Color, Double, String, Bool 
    }
    public abstract class DataEditor
    {
        protected IObjectProperty _property;
        public DataEditor()
        {
        }
        public DataEditor(IObjectProperty value)
        {
            Name = value.Name;
            Property = value;
        }
        public static DataEditor FromVariantValue(IObjectProperty value)
        {
            if (value.Value.NativeType == typeof(int))
            {
                return new IntEditor(value);
            }
            if (value.Value.NativeType == typeof(double))
            {
                return new DoubleEditor(value);
            }
            if (value.Value.NativeType == typeof(Color))
            {
                return new ColorEditor(value);
            }
            if (value.Value.NativeType == typeof(bool))
            {
                return new BoolEditor(value);
            }
            if (value.Value.NativeType == typeof(string))
            {
                return new StringEditor(value);
            }
            return null;
        }
        public static DataEditor FromDataType(DataType dt)
        {
            switch (dt)
            {
                case DataType.Int:
                    return new IntEditor();
                case DataType.Double:
                    return new DoubleEditor();
                case DataType.Bool:
                    return new BoolEditor();
                case DataType.Color:
                    return new ColorEditor();
                case DataType.String:
                    return new StringEditor();
                default:
                    return null;
            }
        }
        [Browsable(false)]
        public static IEnumerable<DataEditor> EditorTypes
        {
            get
            {
                yield return new IntEditor();
                yield return new DoubleEditor();
                yield return new BoolEditor();
                yield return new StringEditor();
                yield return new ColorEditor();
                yield break;
            }
        }
        [Browsable(false)]
        public virtual IObjectProperty Property
        {
            get
            {
                return _property;
            }
            set
            {
                _property = value;
                if (_property != null)
                {
                    Name = _property.Name;
                }
            }
        }
        [Browsable(false)]
        public abstract object DefaultValue { get; }
        [CategoryGlobal("CAT_PROPERTY")]
        [DisplayNameGlobal("DIS_NAME")]
        [DescriptionGlobal("DES_NAME")]
        public string Name { get; set; }
        [CategoryGlobal("CAT_DATA")]
        [DisplayNameGlobal("DIS_TYPE")]
        [DescriptionGlobal("DES_TYPE")]
        public DataType DataType { get; set; }
        [Browsable(false)]
        public abstract object UValue { get; }
    }
    public class IntEditor : DataEditor
    {
        public IntEditor() : base()
        {
            DataType = DataType.Int; 
        }
        public IntEditor(IObjectProperty value) : base(value)
        {
            DataType = DataType.Int;
            Value = value.Value.AsInt;
        }
        public override IObjectProperty Property
        {
            get
            {
                return base.Property;
            }
            set
            {
                base.Property = value;
                if (_property != null)
                {
                    Value = _property.Value.AsInt;
                }
            }
        }
        public override object DefaultValue
        {
            get
            {
                return (int)0;
            }
        }
        [CategoryGlobal("CAT_DATA")]
        [DisplayNameGlobal("DIS_VALUE")]
        [DescriptionGlobal("DES_VALUE")]
        public int Value { get; set; }
        public override object UValue
        {
            get
            {
                return Value;
            }
        }
        public override string ToString()
        {
            return "Int";
        }
    }
    public class ColorEditor : DataEditor
    {
        public ColorEditor() : base()
        {
            DataType = DataType.Color;
        }
        public ColorEditor(IObjectProperty value) : base(value)
        {
            DataType = DataType.Color;
            Value = value.Value.AsColor;
        }
        public override IObjectProperty Property
        {
            get
            {
                return base.Property;
            }
            set
            {
                base.Property = value;
                if (_property != null)
                {
                    Value = _property.Value.AsColor;
                }
            }
        }
        public override object DefaultValue
        {
            get
            {
                return Color.Black;
            }
        }
        [CategoryGlobal("CAT_DATA")]
        [DisplayNameGlobal("DIS_VALUE")]
        [DescriptionGlobal("DES_VALUE")]
        public Color Value { get; set; }
        public override object UValue
        {
            get
            {
                return Value;
            }
        }
        public override string ToString()
        {
            return "Color";
        }
    }
    public class DoubleEditor : DataEditor
    {
        public DoubleEditor() : base()
        {
            DataType = DataType.Double;
        }
        public DoubleEditor(IObjectProperty value) : base(value)
        {
            DataType = DataType.Double;
            Value = value.Value.AsDouble;
        }
        public override IObjectProperty Property
        {
            get
            {
                return base.Property;
            }
            set
            {
                base.Property = value;
                if (_property != null)
                {
                    Value = _property.Value.AsDouble;
                }
            }
        }
        public override object DefaultValue
        {
            get
            {
                return (double)0;
            }
        }
        [CategoryGlobal("CAT_DATA")]
        [DisplayNameGlobal("DIS_VALUE")]
        [DescriptionGlobal("DES_VALUE")]
        public double Value { get; set; }
        public override object UValue
        {
            get
            {
                return Value;
            }
        }
        public override string ToString()
        {
            return "Double";
        }
    }
    public class StringEditor : DataEditor
    {
        public StringEditor() : base()
        {
            DataType = DataType.String;
        }
        public StringEditor(IObjectProperty value) : base(value)
        {
            DataType = DataType.String;
            Value = value.Value.AsString;
        }
        public override IObjectProperty Property
        {
            get
            {
                return base.Property;
            }
            set
            {
                base.Property = value;
                if (_property != null)
                {
                    Value = _property.Value.AsString;
                }
            }
        }
        public override object DefaultValue
        {
            get
            {
                return "";
            }
        }
        [CategoryGlobal("CAT_DATA")]
        [DisplayNameGlobal("DIS_VALUE")]
        [DescriptionGlobal("DES_VALUE")]
        public string Value { get; set; }
        public override object UValue
        {
            get
            {
                return Value;
            }
        }
        public override string ToString()
        {
            return "String";
        }
    }
    public class BoolEditor : DataEditor
    {
        public BoolEditor() : base()
        {
        }
        public BoolEditor(IObjectProperty value) : base(value)
        {
            DataType = DataType.Bool;
            Value = value.Value.AsBool;
        }
        public override IObjectProperty Property
        {
            get
            {
                return base.Property;
            }
            set
            {
                base.Property = value;
                if (_property != null)
                {
                    Value = _property.Value.AsBool;
                }
            }
        }
        public override object DefaultValue
        {
            get
            {
                return false;
            }
        }
        [CategoryGlobal("CAT_DATA")]
        [DisplayNameGlobal("DIS_VALUE")]
        [DescriptionGlobal("DES_VALUE")]
        public bool Value { get; set; }
        public override object UValue
        {
            get
            {
                return Value;
            }
        }
        public override string ToString()
        {
            return "Bool";
        }
    }
}
